# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Processing workflow for single data source processing."""

from os import getenv, getpid
from os.path import basename, exists
import logging
from datetime import timedelta
import inspect
import xarray

# Internal utilities
from geoips.errors import PluginError
from geoips.errors import OutputFormatterDatelineError
from geoips.errors import OutputFormatterInvalidProjectionError
from geoips.filenames.duplicate_files import remove_duplicates
from geoips.geoips_utils import copy_standard_metadata, output_process_times
from geoips.utils.memusg import PidLog
from geoips.xarray_utils.data import sector_xarrays
from geoips.sector_utils.utils import filter_area_defs_actual_time, is_dynamic_sector
from geoips.geoips_utils import replace_geoips_paths

# Old interfaces (YAML, not updated to classes yet!)
from geoips.dev.product import (
    get_required_variables,
    get_requested_datasets_for_variables,
    get_covg_from_product,
    get_covg_args_from_product,
)
from geoips.dev.output_config import (
    get_filename_formatters,
    get_filename_formatter_kwargs,
    get_output_formatter,
    get_output_formatter_kwargs,
    get_metadata_filename_formatter,
    get_metadata_filename_formatter_kwargs,
    get_metadata_output_formatter,
    get_metadata_output_formatter_kwargs,
    get_minimum_coverage,
)

# New class-based interfaces
from geoips.interfaces import (
    algorithms,
    colormappers,
    filename_formatters,
    interpolators,
    output_formatters,
    products,
    readers,
    sector_adjusters,
)

# These output families require an input filename list, AND require the returned
# list of products to match what was passed in
OUTPUT_FAMILIES_WITH_OUTFNAMES_ARG = [
    "xrdict_varlist_outfnames_to_outlist",
    "xrdict_area_product_outfnames_to_outlist",
    "image",
    "unprojected",
    "image_overlay",
    "xarray_data",
]
# These output families do NOT take in a list of filenames, and an arbitrary
# list of output products can be returned - there is no expected output file
# list
OUTPUT_FAMILIES_WITH_NO_OUTFNAMES_ARG = [
    "xrdict_area_product_to_outlist",
    "xrdict_to_outlist",
]

FILENAME_FORMATS_WITHOUT_COVG = [
    "xarray_metadata_to_filename",
    "xarray_area_product_to_filename",
]

FILENAME_FORMATS_FOR_XARRAY_DICT_TO_OUTPUT_FORMAT = [
    "xarray_metadata_to_filename",
    "xarray_area_product_to_filename",
]

# Organizing lists at the top of the single source procflow of all of the
# xarray-dict based product families to better track how they flow through
# the procflow.  In the end, there should be one of
# each xarray-dict based family (with/without algorithm, with/without area,
# with/without sectoring for the area-based ones)
# Including lists at the top so we can more easily identify what families
# are used/supported in the procflows, and if we notice duplicate supported
# families we can add them to this list.
PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITHOUT_AREA = [
    "unsectored_xarray_dict_to_output_format",
]
PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITH_ALGORITHM_WITHOUT_AREA = [
    "unsectored_xarray_dict_to_algorithm_to_output_format",
]
PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITH_AREA = [
    "unsectored_xarray_dict_area_to_output_format",
]
PRODUCT_FAMILIES_OF_SECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITH_AREA = [
    "sectored_xarray_dict_to_output_format",
]
PRODUCT_FAMILIES_OF_XARRAY_DICT_WITH_AREA = (
    PRODUCT_FAMILIES_OF_SECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITH_AREA
    + PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITH_AREA
)
PRODUCT_FAMILIES_OF_XARRAY_DICT_WITHOUT_AREA = (
    PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITHOUT_AREA
    + PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITH_ALGORITHM_WITHOUT_AREA
)

PMW_NUM_PIXELS_X = 1400
PMW_NUM_PIXELS_Y = 1400
PMW_PIXEL_SIZE_X = 1000
PMW_PIXEL_SIZE_Y = 1000

LOG = logging.getLogger(__name__)

interface = "procflows"
family = "standard"
name = "single_source"


def output_all_metadata(
    output_dict, output_fnames, metadata_fnames, xarray_obj, area_def=None
):
    """Output all metadata."""
    final_outputs = output_fnames.copy()
    metadata_output_formatter = get_metadata_output_formatter(output_dict)
    metadata_output_formatter_kwargs = get_metadata_output_formatter_kwargs(output_dict)
    for metadata_fname, output_fname in zip(metadata_fnames, output_fnames):
        if metadata_fname is not None:
            # Optional arguments for standard metadata formats (like "output_dict")
            metadata_output_formatter_kwargs["metadata_fname_dict"] = metadata_fnames[
                metadata_fname
            ]
            metadata_output_formatter_kwargs["output_fname_dict"] = output_fnames[
                output_fname
            ]
            metadata_output_formatter_kwargs["output_dict"] = output_dict
            output_plugin = output_formatters.get_plugin(metadata_output_formatter)
            output_kwargs = remove_unsupported_kwargs(
                output_plugin, metadata_output_formatter_kwargs
            )
            if output_plugin.family == "standard_metadata":
                curr_outputs = output_plugin(
                    area_def,
                    xarray_obj=xarray_obj,
                    metadata_yaml_filename=metadata_fname,
                    product_filename=output_fname,
                    **output_kwargs,
                )
                if curr_outputs != [metadata_fname]:
                    raise ValueError("Did not produce expected products")

                for curr_output in curr_outputs:
                    final_outputs[curr_output] = metadata_fnames[curr_output]

    return final_outputs


def get_output_filenames(
    fname_formats,
    output_dict,
    prod_plugin,
    xarray_obj=None,
    area_def=None,
    supported_filenamer_types=None,
):
    """Get output filenames."""
    output_fnames = {}
    metadata_fnames = {}
    for filename_formatter in fname_formats:
        filename_formatter_kwargs = get_filename_formatter_kwargs(
            filename_formatter, output_dict
        )
        metadata_filename_formatter = get_metadata_filename_formatter(
            filename_formatter, output_dict
        )
        metadata_filename_formatter_kwargs = get_metadata_filename_formatter_kwargs(
            metadata_filename_formatter, output_dict
        )

        output_fname = get_filename(
            filename_formatter,
            prod_plugin,
            xarray_obj,
            area_def,
            output_dict=output_dict,
            supported_filenamer_types=supported_filenamer_types,
            filename_formatter_kwargs=filename_formatter_kwargs,
        )

        # If we weren't able to get a valid output filename, do not proceed.
        if output_fname is None:
            continue

        output_fnames[output_fname] = {
            "filename_formatter": filename_formatter,
            "filename_formatter_kwargs": filename_formatter_kwargs,
            "product_name": prod_plugin.name,
        }

        metadata_fname = None
        if metadata_filename_formatter:
            fname_fmt_plugin = filename_formatters.get_plugin(
                metadata_filename_formatter
            )
            if fname_fmt_plugin.family == "standard_metadata":
                metadata_filename_formatter_kwargs = remove_unsupported_kwargs(
                    fname_fmt_plugin, metadata_filename_formatter_kwargs
                )
                metadata_fname = fname_fmt_plugin(
                    area_def,
                    xarray_obj,
                    output_fname,
                    **metadata_filename_formatter_kwargs,
                )
        metadata_fnames[metadata_fname] = {
            "filename_formatter": filename_formatter,
            "filename_formatter_kwargs": filename_formatter_kwargs,
            "metadata_filename_formatter": metadata_filename_formatter,
            "metadata_filename_formatter_kwargs": metadata_filename_formatter_kwargs,
            "product_name": prod_plugin.name,
        }
    return output_fnames, metadata_fnames


def add_attrs_from_area_def(final_xarray, source_xarray, area_def):
    """Add attributes from an area_def."""
    # MLS I think this should actually just be final_xarray and final_xarray,
    # no source_xarray!  We might be losing information...
    # Ensure we have the "adjustment"id" in the filename appropriately
    if area_def and "adjustment_id" in area_def.sector_info:
        final_xarray = add_filename_extra_field(
            source_xarray, "adjustment_id", area_def.sector_info["adjustment_id"]
        )
    return final_xarray


def resector_xarrays(
    resector, sect_xarrays, area_def, variables, window_start_time, window_end_time
):
    """Resector xarrays if requested."""
    # If the initial sectoring was to a padded area definition, must sector to final
    # area_def here.
    # Allow specifying whether it needs to be resectored or not via kwargs.
    if resector:
        LOG.interactive("Resectoring xarrays without padding...")
        curr_sect_xarrays = sector_xarrays(
            sect_xarrays,
            area_def,
            varlist=variables,
            hours_before_sector_time=6,
            hours_after_sector_time=9,
            drop=True,
            window_start_time=window_start_time,
            window_end_time=window_end_time,
        )
        # hours_before_sector_time=6, hours_after_sector_time=6, drop=True)
    else:
        curr_sect_xarrays = sect_xarrays
    return curr_sect_xarrays


def get_interp_plugin_from_product(prod_plugin):
    """Get the interpolator plugin from the product spec.

    Reassign interp_plugin based on CURRENT sect_xarray
    Allow re-defining interpolation for different datasets.
    """
    interp_plugin = None
    if "interpolator" in prod_plugin["spec"]:
        interp_plugin = interpolators.get_plugin(
            prod_plugin["spec"]["interpolator"]["plugin"]["name"]
        )
        interp_args = prod_plugin["spec"]["interpolator"]["plugin"]["arguments"]
        interp_args = remove_unsupported_kwargs(interp_plugin, interp_args)
    return interp_plugin, interp_args


def apply_interp_first(
    variables,
    curr_sect_xarrays,
    prod_plugin,
    datasets_for_vars,
    resampled_read,
    area_def,
    processed_xarrays,
):
    """Apply interpolation first.

    For product types that involve interpolation before algorithm.
    """
    # Default to empty xarray.Dataset() - will be populated within loop with
    # appropriate regridded variables.
    interp_xarray = xarray.Dataset()
    for varname in variables:
        LOG.info("TRYING variable %s", varname)
        for key, sect_xarray in curr_sect_xarrays.items():
            LOG.info("    TRYING dataset %s for variable %s", key, varname)

            if varname not in sect_xarray.variables:
                continue

            # Determine which interpolator to use based on the product definition.
            interp_plugin, interp_args = get_interp_plugin_from_product(prod_plugin)

            # Check if a specific dataset was specified for the current variable,
            # and ensure we are pulling the variable from the correct dataset.
            if not use_variable_from_current_dataset(
                varname,
                key,
                variables,
                sect_xarray,
                interp_xarray,
                resampled_read,
                datasets_for_vars,
            ):
                continue

            # Potential efficiency hit with to_masked_array for dask arrays, etc
            # LOG.info('Min/max %s %s / %s, dataset %s',
            #          varname,
            #          sect_xarray[varname].to_masked_array().min(),
            #          sect_xarray[varname].to_masked_array().max(),
            #          key)

            # It is much faster to interpolate all variables at once than
            # one at a time. Since we can assume "variables" is all of the
            # requested variables, just interpolate them all the first time
            # through, and add logic to NOT reinterpolate variables if they've
            # already been interpolated.
            # For reference, with 11 variables, it was 1 min 13s
            # to interpolate individually, 22s in one call.
            interp_args["varlist"] = variables
            if "time_dim" in sect_xarray.dims:
                # This is for a particularly formatted dataset, that includes
                # separate arrays for different times (ABI fire product).
                # We need to be careful this does not break for some other
                # dataset that includes a differently formatted "time"
                # dimension.
                tdims = len(sect_xarray.time_dim)

                interp_list = []
                for i in range(tdims):
                    interp_list.append(
                        perform_interpolation(
                            interp_plugin,
                            area_def,
                            sect_xarray.isel(time_dim=i),
                            xarray.Dataset(),
                            interp_args,
                            processed_xarrays,
                        )
                    )

                interp_xarray[varname] = xarray.concat(interp_list, dim="dim_2")[
                    varname
                ]
            else:
                interp_xarray = perform_interpolation(
                    interp_plugin,
                    area_def,
                    sect_xarray,
                    interp_xarray,
                    interp_args,
                    processed_xarrays,
                )

            # Potential efficiency hit with to_masked_array for dask arrays, etc
            # LOG.info('Min/max interp %s %s / %s',
            #          varname,
            #          interp_xarray[varname].to_masked_array().min(),
            #          interp_xarray[varname].to_masked_array().max())
    return interp_xarray


def use_variable_from_current_dataset(
    varname,
    key,
    variables,
    sect_xarray,
    interp_xarray,
    resampled_read,
    datasets_for_vars,
):
    """Use the variable from the current dataset.

    If a specific dataset was requested for the current variable, and
    this dataset was NOT requested via a resampled_read (in which case
    the native datasets won't exist, only the resampled dataset),
    then use the appropriately requested dataset.
    """
    if varname in datasets_for_vars and not resampled_read:
        if key in datasets_for_vars[varname]:
            LOG.info(
                "        USING %s varname from dataset %s, as specified in "
                "product_input YAML config",
                varname,
                key,
            )
        else:
            LOG.info(
                "        WAITING dataset %s not requested for variable %s in "
                "product_input YAML config",
                key,
                varname,
            )
            return False
    # If we've already interpolated this variable, check if it is needed
    # before interpolating again
    elif interp_xarray is not None and varname in list(interp_xarray.keys()):
        # If all of the required variables are in the current dataset, use this
        # version
        if set(variables).issubset(set(sect_xarray.variables.keys())):
            LOG.info(
                "        REPLACING %s with current dataset %s, all required "
                "variables in current dataset",
                varname,
                key,
            )
        # Otherwise, skip re-interpolating to avoid unecessary computation
        else:
            LOG.warning(
                "        SKIPPING %s, encountered multiple versions, skipping "
                "subsequent dataset %s",
                varname,
                key,
            )
            return False
    else:
        LOG.info(
            "        USING %s varname from dataset %s - first availalbe, and "
            "not specified in YAML",
            varname,
            key,
        )
    return True


def get_alg_and_interp_plugins(prod_plugin):
    """Get algorithm and interpolator plugins from prod_plugin definition."""
    # Only attempt to set algorithm function if algorithm requested in product type
    try:
        alg_args = prod_plugin["spec"]["algorithm"]["plugin"]["arguments"]
        alg_plugin = algorithms.get_plugin(
            prod_plugin["spec"]["algorithm"]["plugin"]["name"]
        )
    except KeyError:
        alg_args = None
        alg_plugin = None

    try:
        interp_args = prod_plugin["spec"]["interpolator"]["plugin"]["arguments"]
        interp_plugin = interpolators.get_plugin(
            prod_plugin["spec"]["interpolator"]["plugin"]["name"]
        )
        interp_args = remove_unsupported_kwargs(interp_plugin, interp_args)
    except KeyError:
        interp_args = None
        interp_plugin = None
    return alg_plugin, alg_args, interp_plugin, interp_args


def apply_alg_after_interp(
    interp_xarray,
    area_def,
    alg_plugin,
    alg_args,
    prod_plugin,
    variables,
    processed_xarrays,
):
    """Apply algorithm after interpolation.

    MLS need to add ability here to pull from processed_xarrays if algorithm
    was already applied.
    """
    if processed_xarrays and prod_plugin.name in processed_xarrays:
        return processed_xarrays[prod_plugin.name]
    # Specify the call signature and return value for different algorithm types:
    if prod_plugin.family in ["interpolator"]:
        # Note "interp" product type will NOT have a single variable named
        # "product_name", just the individual interpolated variables.
        interp_xarray = interp_xarray
    elif alg_plugin.family in ["xarray_to_numpy"]:
        # xarray_to_numpy will return a single array, which can be set to the
        # "product_name" variable.
        LOG.interactive(
            "  Applying '%s' family algorithm '%s' to data...",
            alg_plugin.family,
            alg_plugin.name,
        )
        interp_xarray[prod_plugin.name] = xarray.DataArray(
            alg_plugin(interp_xarray, **alg_args)
        )
    elif alg_plugin.family in ["xarray_to_xarray"]:
        # xarray_to_xarray algorithm type will return the full xarray object -
        # assume variable names have been set appropriately within the
        # algorithm.  This could be another good use of the
        # "alt_varname_for_covg" kwarg in the coverage checks - if we want to
        # just use a specific variable for the coverage checks rather than the
        # "product_name" variable.
        LOG.interactive(
            "  Applying '%s' family algorithm '%s' to data...",
            alg_plugin.family,
            alg_plugin.name,
        )
        interp_xarray = alg_plugin(
            interp_xarray, variables, prod_plugin.name, **alg_args
        )
    elif alg_plugin.family in [
        "single_channel",
        "channel_combination",
        "list_numpy_to_numpy",
        "rgb",
    ]:
        # Assume ANYTHING else takes in a list of numpy arrays, and returns a
        # single numpy array.
        # Perhaps we should be explicit here...
        LOG.interactive(
            "  Applying '%s' family algorithm '%s' to data...",
            alg_plugin.family,
            alg_plugin.name,
        )
        interp_xarray[prod_plugin.name] = xarray.DataArray(
            alg_plugin(
                [interp_xarray[varname].to_masked_array() for varname in variables],
                **alg_args,
            )
        )
    else:
        raise TypeError(
            f"UNSUPPORTED algorithm family '{alg_plugin.family}' or product family "
            f"'{prod_plugin.family}', please add to the single_source procflow's "
            "'get_alg_xarray' function appropriately"
        )
    return interp_xarray


def apply_alg_first(
    alg_plugin,
    alg_args,
    prod_plugin,
    curr_sect_xarrays,
    sect_xarrays,
    variables,
    variable_names,
    area_def,
):
    """Apply algorithm appropriately based on algorithm family.

    MLS
    Inexplicably some of these use curr_sect_xarrays, and some use sect_xarrays.
    Also, some use variables and some use variable_names.
    I am guessing there is no reason for the difference, but maintaining the
    original functionality for now.
    """
    alg_xarray = xarray.Dataset()
    alg_xarray.attrs = sect_xarrays["METADATA"].attrs.copy()
    if alg_plugin.family in ["xarray_to_numpy"]:
        # Why does this one use sect_xarrays and not curr_sect_xarrays?
        # And it uses variable_names, not variables.
        # print(variable_names,variables)
        alg_xarray = apply_alg_xarray_to_numpy(
            alg_xarray, alg_plugin, alg_args, prod_plugin, sect_xarrays, variables
        )
    elif alg_plugin.family in ["xarray_dict_area_def_to_numpy"]:
        # Why does this one use sect_xarrays and not curr_sect_xarrays?
        alg_xarray = apply_alg_xarray_dict_area_def_to_numpy(
            alg_xarray, alg_plugin, alg_args, prod_plugin, sect_xarrays, area_def
        )
    elif alg_plugin.family in ["xarray_dict_to_xarray"]:
        # This one uses sect_xarrays
        alg_xarray = apply_alg_xarray_dict_to_xarray(alg_plugin, alg_args, sect_xarrays)
    elif alg_plugin.family in ["xarray_dict_to_xarray_dict"]:
        # This one uses sect_xarrays
        alg_xarray = apply_alg_xarray_dict_to_xarray_dict(
            alg_plugin, alg_args, sect_xarrays
        )
    elif alg_plugin.family in ["xarray_to_xarray"]:
        # This one uses variables, not variable_names
        alg_xarray = apply_alg_xarray_to_xarray(
            alg_plugin, alg_args, prod_plugin, sect_xarrays, variables
        )
    elif alg_plugin.family in ["list_numpy_to_numpy"]:
        # Why does this one use curr_sect_xarrays and not sect_xarrays?
        # And variables, not variable_names ?
        alg_xarray = apply_alg_list_numpy_to_numpy(
            alg_xarray,
            alg_plugin,
            alg_args,
            prod_plugin,
            curr_sect_xarrays,
            variables,
        )
    return alg_xarray


def apply_interp_after_alg(
    alg_xarray, interp_plugin, interp_args, prod_plugin, area_def, processed_xarrays
):
    """Apply interpolation after algorithm."""
    # No interpolation required
    if prod_plugin.family in ["algorithm", "algorithm_colormapper"]:
        final_xarray = alg_xarray
    # If required, interpolate the result prior to returning
    elif prod_plugin.family == "algorithm_interpolator_colormapper":
        if interp_args.get("varlist") is None or interp_args["varlist"] is None:
            # Just assume the only thing we're interpolating is the product name itself
            interp_args["varlist"] = [prod_plugin.name]
        # Otherwise use the varlist that was provided. For example, a user could produce
        # Additional variables from their algorithm alongside their product name
        final_xarray = perform_interpolation(
            interp_plugin,
            area_def,
            alg_xarray,
            xarray.Dataset(),
            interp_args,
            processed_xarrays,
        )
    return final_xarray


def apply_alg_xarray_to_xarray(
    alg_plugin, alg_args, prod_plugin, sect_xarrays, variable_names
):
    """Apply xarray_to_xarray algorithm."""
    input_alg_xarray = None
    for varname in variable_names:
        LOG.info("TRYING variable %s for non-interpolated algorithms", varname)
        for curr_sect_xarray in sect_xarrays.values():
            if varname in curr_sect_xarray:
                if input_alg_xarray is None:
                    LOG.info(
                        "    USING sectored xarray %s for non-interpolated "
                        "algorithms",
                        curr_sect_xarray,
                    )
                    input_alg_xarray = curr_sect_xarray
                else:
                    LOG.info(
                        "    SKIPPING For non-interpolated data processing, "
                        "all native variables must be the same resolution! "
                        "Skipping variable %s, shape %s, input_alg_xarrays: %s",
                        varname,
                        curr_sect_xarray[varname].shape,
                        input_alg_xarray,
                    )
    if input_alg_xarray is None:
        raise ValueError(
            "No required variables in any xarrays for 'xarray_to_xarray' "
            "algorithm type"
        )
    LOG.interactive(
        "  Applying '%s' family algorithm '%s' to data...",
        alg_plugin.family,
        alg_plugin.name,
    )
    alg_xarray = alg_plugin(
        input_alg_xarray, variable_names, prod_plugin.name, **alg_args
    )
    return alg_xarray


def apply_alg_xarray_dict_to_xarray(alg_plugin, alg_args, sect_xarrays):
    """Apply xarray_dict_to_xarray algorithm."""
    # Format the call signature for passing a dictionary of xarrays,
    # plus area_def, and return a single numpy array
    LOG.interactive(
        "  Applying '%s' family algorithm '%s' to data...",
        alg_plugin.family,
        alg_plugin.name,
    )
    alg_xarray = alg_plugin(sect_xarrays, **alg_args)
    return alg_xarray


def apply_alg_xarray_dict_to_xarray_dict(alg_plugin, alg_args, sect_xarrays):
    """Apply xarray_dict_to_xarray algorithm."""
    # Format the call signature for passing a dictionary of xarrays,
    # plus area_def, and return a single numpy array
    LOG.interactive(
        "  Applying '%s' family algorithm '%s' to data...",
        alg_plugin.family,
        alg_plugin.name,
    )
    alg_xarray_dict = alg_plugin(sect_xarrays, **alg_args)
    return alg_xarray_dict


def apply_alg_xarray_dict_area_def_to_numpy(
    alg_xarray, alg_plugin, alg_args, prod_plugin, sect_xarrays, area_def
):
    """Apply xarray_dict_area_def_to_numpy algorithm."""
    # Format the call signature for passing a dictionary of xarrays,
    # plus area_def, and return a single numpy array
    LOG.interactive(
        "  Applying '%s' family algorithm '%s' to data...",
        alg_plugin.family,
        alg_plugin.name,
    )
    alg_xarray[prod_plugin.name] = xarray.DataArray(
        alg_plugin(sect_xarrays, area_def, **alg_args)
    )
    return alg_xarray


def apply_alg_xarray_to_numpy(
    alg_xarray, alg_plugin, alg_args, prod_plugin, sect_xarrays, variable_names
):
    """Apply xarray_to_numpy algorithm."""
    # Format the call signature for passing a dictionary of xarrays,
    # plus area_def, and return a single numpy array
    for dsname in sect_xarrays.keys():
        if set(variable_names).issubset(set(sect_xarrays[dsname].variables.keys())):
            LOG.interactive(
                "  Applying '%s' family algorithm '%s' to data...",
                alg_plugin.family,
                alg_plugin.name,
            )
            alg_xarray = sect_xarrays[dsname]
            alg_xarray[prod_plugin.name] = xarray.DataArray(
                alg_plugin(sect_xarrays[dsname], **alg_args)
            )
            # drops geo values?
    return alg_xarray


def apply_alg_list_numpy_to_numpy(
    alg_xarray, alg_plugin, alg_args, prod_plugin, sect_xarrays, variables
):
    """Apply list_numpy_to_numpy algorithm."""
    # Need to pull all the required variables out of the various xarray datasets
    # and add them to numpy list.
    # Then assign the resulting numpy array to the "product_name" DataArray
    # within the xarray Dataset
    numpys = []
    for varname in variables:
        for curr_sect_xarray in sect_xarrays.values():
            if varname in list(curr_sect_xarray.variables.keys()):
                numpys += [curr_sect_xarray[varname].to_masked_array()]
                alg_xarray = curr_sect_xarray
    LOG.interactive(
        "  Applying '%s' family algorithm '%s' to data...",
        alg_plugin.family,
        alg_plugin.name,
    )
    alg_xarray[prod_plugin.name] = xarray.DataArray(alg_plugin(numpys, **alg_args))
    return alg_xarray


def perform_interpolation(
    interp_plugin, area_def, sect_xarray, interp_xarray, interp_args, processed_xarrays
):
    """Perform standard interpolation."""
    interp_args, interp_xarray = select_variables_to_interp(
        interp_args, area_def, sect_xarray, interp_xarray, processed_xarrays
    )
    LOG.interactive(
        "  Interpolating data with interpolator '%s' args '%s'...",
        interp_plugin.name,
        interp_args,
    )
    # Only attempt to interpolate if there are required variables left.
    if interp_args["varlist"]:
        interp_xarray = interp_plugin(
            area_def, sect_xarray, interp_xarray, **interp_args
        )
    else:
        LOG.info("No variables to interpolate, returning interp_xarray unchanged.")
    return interp_xarray


def get_unique_dataset_key(area_def, xobj):
    """Get a unique id for xarray dataset."""
    standard_attrs = (
        f"{xobj.attrs['start_datetime']}_"
        f"{xobj.attrs['end_datetime']}_"
        f"{xobj.attrs['platform_name']}_"
        f"{xobj.attrs['source_name']}"
    )
    return f"{area_def.area_id}_{hash(area_def)}_{standard_attrs}"


def select_variables_to_interp(
    interp_args, area_def, source_xarray, interp_xarray, processed_xarrays
):
    """Select interpolation variables."""
    curr_available_vars = set(list(source_xarray.variables.keys()))
    already_interped_vars = set(list(interp_xarray.variables.keys()))
    all_requested_vars = set(interp_args["varlist"])
    curr_requested_vars = curr_available_vars.intersection(all_requested_vars)
    interp_vars = curr_requested_vars.difference(already_interped_vars)
    final_interp_vars = list(interp_vars)
    # This is a unique identifier so we can re-use variables
    # and product arrays appropriately.
    area_def_key = get_unique_dataset_key(area_def, source_xarray)
    if processed_xarrays and area_def_key in processed_xarrays:
        LOG.info(
            "  Area def '%s' in processed_xarrays, "
            "with variables '%s', using existing array, "
            "in place of interp_xarray '%s'.",
            area_def_key,
            list(processed_xarrays[area_def_key].variables.keys()),
            interp_xarray,
        )
        # Just add to this existing array
        interp_xarray = processed_xarrays[area_def_key]
        final_interp_vars = []
        for varname in interp_vars:
            if varname not in processed_xarrays[area_def_key].variables.keys():
                final_interp_vars += [varname]
    interp_args["varlist"] = final_interp_vars
    return interp_args, interp_xarray


def remove_unsupported_kwargs(module, requested_kwargs):
    """Remove unsupported keyword arguments."""
    module_args = set(inspect.signature(module).parameters.keys())
    unsupported = list(set(requested_kwargs.keys()).difference(module_args))
    if "kwargs" not in module_args:
        for key in unsupported:
            LOG.warning("REMOVING UNSUPPORTED %s key %s", module, key)
            requested_kwargs.pop(key)
    return requested_kwargs


def add_filename_extra_field(xarray_obj, field_name, field_value):
    """Add filename extra field."""
    if "filename_extra_fields" not in xarray_obj.attrs:
        xarray_obj.attrs["filename_extra_fields"] = {}
    xarray_obj.attrs["filename_extra_fields"][field_name] = field_value
    return xarray_obj


def combine_filename_extra_fields(source_xarray, dest_xarray):
    """Combine filename extra fields."""
    if "filename_extra_fields" in source_xarray.attrs:
        for field in source_xarray.filename_extra_fields:
            if "filename_extra_fields" not in dest_xarray.attrs:
                dest_xarray.attrs["filename_extra_fields"] = {}
            dest_xarray.attrs["filename_extra_fields"][field] = (
                source_xarray.filename_extra_fields[field]
            )
    return dest_xarray


def process_sectored_data_output(
    xobjs, variables, prod_plugin, output_dict, area_def=None
):
    """Process sectored data output.

    If current product family requires a sectored dictionary of xarrays, does
    not apply an algorithm, and DOES require an area definition, call
    'process_xarray_dict_to_output_format', store the result in a list, and return it.
    """
    output_products = []
    # Currently this only supports NOT applying an algorithm.  Could expand
    # to additionally support algorithm application.
    if (
        prod_plugin.family
        in PRODUCT_FAMILIES_OF_SECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITH_AREA
    ):
        output_products += process_xarray_dict_to_output_format(
            xobjs, variables, prod_plugin, output_dict, area_def=area_def
        )
    return output_products


def process_xarray_dict_to_output_format(
    xobjs, variables, prod_plugin, output_dict, area_def=None
):
    """Process xarray dict to output format."""
    output_formatter = get_output_formatter(output_dict)
    output_formatter_kwargs = get_output_formatter_kwargs(output_dict)
    # All xarray dict based products can be processed here - this function is
    # called from both within the loop through all area_defs, and before it.
    # As well as for both sectored / non-sectored products.
    supported_product_types = (
        PRODUCT_FAMILIES_OF_XARRAY_DICT_WITH_AREA
        + PRODUCT_FAMILIES_OF_XARRAY_DICT_WITHOUT_AREA
    )

    if prod_plugin.family not in supported_product_types:
        raise TypeError(
            f"UNSUPPORTED product_type {prod_plugin.family} "
            f'for product {prod_plugin.name} source {xobjs["METADATA"].source_name} \n'
            f"      product_type must be one of {supported_product_types}"
        )

    # These are all the supported oututter familes
    supported_output_plugin_types = (
        OUTPUT_FAMILIES_WITH_OUTFNAMES_ARG + OUTPUT_FAMILIES_WITH_NO_OUTFNAMES_ARG
    )

    output_plugin = output_formatters.get_plugin(output_formatter)
    # Default to empty dict for output_fnames
    output_fnames = {}

    # Only get output filenames if needed
    if output_plugin.family in OUTPUT_FAMILIES_WITH_OUTFNAMES_ARG:
        supported_filenamer_types = FILENAME_FORMATS_FOR_XARRAY_DICT_TO_OUTPUT_FORMAT
        fname_formats = get_filename_formatters(output_dict)
        output_fnames, metadata_fnames = get_output_filenames(
            fname_formats,
            output_dict,
            prod_plugin,
            xarray_obj=xobjs["METADATA"],
            area_def=area_def,
            supported_filenamer_types=supported_filenamer_types,
        )
        LOG.interactive(
            "Output filenames for family %s: %s",
            output_plugin.family,
            output_fnames,
        )
    else:
        LOG.interactive(
            "No output filenames required for family %s", output_plugin.family
        )

    if "output_dict" not in output_formatter_kwargs:
        output_formatter_kwargs["output_dict"] = output_dict
    output_formatter_kwargs = remove_unsupported_kwargs(
        output_plugin, output_formatter_kwargs
    )

    LOG.interactive("Running output plugin of family %s", output_plugin.family)
    if output_plugin.family == "xrdict_varlist_outfnames_to_outlist":
        curr_products = output_plugin(
            xobjs, variables, list(output_fnames.keys()), **output_formatter_kwargs
        )
        # If we pass it a list of filenames, assume we create exactly those filenames
        if curr_products != list(output_fnames.keys()):
            raise (ValueError("Did not produce expected products"))

    elif output_plugin.family == "xrdict_area_product_outfnames_to_outlist":
        curr_products = output_plugin(
            xobjs,
            area_def,
            prod_plugin,
            list(output_fnames.keys()),
            **output_formatter_kwargs,
        )
        # If we pass it a list of filenames, assume we create exactly those filenames
        if curr_products != list(output_fnames.keys()):
            raise (ValueError("Did not produce expected products"))

    elif output_plugin.family == "xrdict_area_product_to_outlist":
        curr_products = output_plugin(
            xobjs, area_def, prod_plugin, **output_formatter_kwargs
        )
        # No input filename list, no check that output filename list matches
        LOG.info(
            "Not checking output file list for output family %s", output_plugin.family
        )
    elif output_plugin.family == "xrdict_to_outlist":
        curr_products = output_plugin(xobjs, **output_formatter_kwargs)
        # No input filename list, no check that output filename list matches
        LOG.info(
            "Not checking output file list for output family %s", output_plugin.family
        )

    # We theoretically can do this for all output_formatters that don't require an area
    # definition
    elif output_plugin.family == "unprojected":
        # If there is not a colormap dictionary already provided in output_formatter
        # kwargs, then try to retrieve it here. If it's not possible, raise a
        # PluginError which points out that no colormapper plugin was provided and
        # therefore we can't produce an unprojected image.
        if "mpl_colors_info" not in output_formatter_kwargs:
            prd_plg = products.get_plugin(
                prod_plugin["source_names"][0], prod_plugin.name
            )
            cmap_name = (
                prd_plg["spec"].get("colormapper", {}).get("plugin", {}).get("name")
            )
            if cmap_name:
                cmap_plg = colormappers.get_plugin(cmap_name)
                output_formatter_kwargs["mpl_colors_info"] = cmap_plg(
                    **prd_plg["spec"]["colormapper"]["plugin"]["arguments"]
                )
        else:
            raise PluginError(
                f"Error: product plugin '{prod_plugin.name}' does not have a "
                "colormapper plugin that is needed for the 'unprojected_image'"
                "output formatter. Please add a colormapper plugin to your product, and"
                " try again."
            )
        # Unprojected Output formatter expects the xarray including the data for your
        # product rather than a dictionary containing that xarray. We probably could
        # just default to 'xobjs[prod_plugin.name]', however adding this conditional
        # doesn't hurt in the case xobjs is an xarray.Dataset(). Shouldn't happen,
        # but again, doesn't hurt to add this functionality.
        if isinstance(xobjs, dict):
            in_xobjs = xobjs[prod_plugin.name]
        else:
            in_xobjs = xobjs
        # Apply unprojected image output formatter
        LOG.info("Applying output formatter of family %s", output_plugin.family)
        curr_products = output_plugin(
            in_xobjs,
            prod_plugin.name,
            output_fnames,
            **output_formatter_kwargs,
        )

    else:
        raise TypeError(
            f'UNSUPPORTED output_formatter "{output_formatter}" '
            f"for product_family {prod_plugin.family}\n"
            f'      output_plugin_family: "{output_plugin.family}"\n'
            f"      output_plugin_type must be one of {supported_output_plugin_types}"
        )

    # We only pre-generated metadata filenames if we also pre-generated output
    # product filenames
    if output_plugin.family in OUTPUT_FAMILIES_WITH_OUTFNAMES_ARG:
        LOG.interactive(
            "Writing out all metadata for products of family %s", output_plugin.family
        )
        final_products = output_all_metadata(
            output_dict,
            output_fnames,
            metadata_fnames,
            xobjs["METADATA"],
            area_def=area_def,
        )
    else:
        final_products = curr_products

    return final_products


def pad_area_definition(
    area_def, source_name=None, force_pad=False, x_scale_factor=1.5, y_scale_factor=1.5
):
    """Pad area definition."""
    from geoips.sector_utils.utils import is_sector_type

    # Always pad TC sectors, and if "force_pad=True" is passed into the function
    if is_sector_type(area_def, "tc") or force_pad:
        LOG.info("Trying area_def %s, %s", area_def.name, area_def.sector_info)
        # Get an extra 50% size for TCs so we can handle recentering and not have
        # missing data. --larger area for possibly moved center for vis/ir backgrounds
        # Default to 1.5x padding
        num_lines = int(area_def.y_size * y_scale_factor)
        num_samples = int(area_def.x_size * x_scale_factor)
        # Need full swath width for AMSU-B and MHS. Need a better solution for this.
        if source_name is not None and source_name in ["amsu-b", "mhs"]:
            num_lines = int(area_def.y_size * 1)
            num_samples = int(area_def.x_size * 5)

        # TC sectors have center lat and center lon defined within the sector_info
        # For other sectors, use lat_0 and lon_0 from proj_dict
        # Do not use proj_dict for TC sectors, because we want the center of the
        # storm, not current center of image.
        if is_sector_type(area_def, "tc"):
            clat = area_def.sector_info["clat"]
            clon = area_def.sector_info["clon"]
        else:
            clat = area_def.proj_dict["lat_0"]
            clon = area_def.proj_dict["lon_0"]

        from geoips.plugins.modules.sector_spec_generators.center_coordinates import (
            call,
        )

        pad_area_def = call(
            area_id=area_def.area_id,
            long_description=area_def.description,
            clat=clat,
            clon=clon,
            projection="eqc",
            num_lines=num_lines,
            num_samples=num_samples,
            pixel_width=area_def.pixel_size_x,
            pixel_height=area_def.pixel_size_y,
        )
        from geoips.sector_utils.utils import copy_sector_info

        pad_area_def = copy_sector_info(area_def, pad_area_def)
    else:
        pad_area_def = area_def
    return pad_area_def


def get_filename(
    filename_formatter,
    prod_plugin=None,
    alg_xarray=None,
    area_def=None,
    supported_filenamer_types=None,
    output_dict=None,
    filename_formatter_kwargs=None,
):
    """Get filename."""
    filename_fmt_plugin = filename_formatters.get_plugin(filename_formatter)
    if (
        supported_filenamer_types is not None
        and filename_fmt_plugin.family not in supported_filenamer_types
    ):
        raise TypeError(
            f'UNSUPPORTED filename_formatter "{filename_formatter}" '
            f'      filenamer_type: "{filename_fmt_plugin.family}"\n'
            f"      filenamer_type must be one of {supported_filenamer_types}"
        )

    # They all use covg except those in list
    if filename_fmt_plugin.family not in FILENAME_FORMATS_WITHOUT_COVG:
        covg_plugin = get_covg_from_product(
            prod_plugin,
            covg_field="filename_coverage_checker",
        )
        covg_args = get_covg_args_from_product(
            prod_plugin,
            covg_field="filename_coverage_checker",
        )
        # Set variable_name to prod_plugin.name if not defined.
        # Always use variable_name if it is defined.
        # Remove from args so there is not a duplicate are when passing
        # (since we are passing covg_varname explicitly).
        # Note get_covg_args_from_product was updated to return a copy of
        # covg_args, so this does not impact future uses of the product.
        covg_varname = covg_args.pop("variable_name", prod_plugin.name)
        # Note variables can be specified as DATASET:VARIABLE, since this is a
        # preprocessed alg_xarray, and not a dictionary of datasets, just use the
        # variable name (we expect the correct variable will exist in this final
        # processed array)
        if ":" in covg_varname:
            covg_varname = covg_varname.split(":")[1]
        covg = covg_plugin(
            alg_xarray,
            covg_varname,
            area_def,
            **covg_args,
        )

    curr_kwargs = remove_unsupported_kwargs(
        filename_fmt_plugin, filename_formatter_kwargs
    )
    if filename_fmt_plugin.family == "data":
        fname = filename_fmt_plugin(
            area_def,
            alg_xarray,
            [prod_plugin.name, "latitude", "longitude"],
            covg,
            **curr_kwargs,
        )
    elif filename_fmt_plugin.family == "xarray_metadata_to_filename":
        fname = filename_fmt_plugin(alg_xarray, **curr_kwargs)
    elif filename_fmt_plugin.family == "xarray_area_product_to_filename":
        fname = filename_fmt_plugin(
            alg_xarray, area_def, prod_plugin.name, **curr_kwargs
        )
    else:
        fname = filename_fmt_plugin(
            area_def, alg_xarray, prod_plugin.name, covg, **curr_kwargs
        )
    return fname


def plot_data(
    output_dict,
    alg_xarray,
    area_def,
    prod_plugin,
    output_kwargs,
    fused_xarray_dict=None,
    no_output=False,
):
    """Plot data.

    alg_xarray used for filename formats, etc.
    If included, fused_xarray_dict used for output format call
    """
    # If keyword argument is allowed for output function, include it
    output_kwargs["output_dict"] = output_dict
    output_formatter = get_output_formatter(output_dict)
    output_plugin = output_formatters.get_plugin(output_formatter)
    output_kwargs = remove_unsupported_kwargs(output_plugin, output_kwargs)
    exclude_platforms = prod_plugin["spec"].get("exclude_platforms")
    include_platforms = prod_plugin["spec"].get("include_platforms")
    if exclude_platforms and alg_xarray.platform_name in exclude_platforms:
        LOG.interactive(
            "SKIPPING Platform %s not requested for product '%s'"
            "Excluded platforms: '%s'...",
            alg_xarray.platform_name,
            prod_plugin.name,
            exclude_platforms,
        )
        return {}
    if include_platforms and (alg_xarray.platform_name not in include_platforms):
        LOG.interactive(
            "SKIPPING Platform %s not requested for product '%s' "
            "Only included platforms: '%s'...",
            alg_xarray.platform_name,
            prod_plugin.name,
            include_platforms,
        )
        return {}

    LOG.interactive(
        "Producing product '%s' final outputs as type '%s'...",
        prod_plugin.name,
        output_plugin.name,
    )

    if no_output or output_plugin.family in OUTPUT_FAMILIES_WITH_NO_OUTFNAMES_ARG:
        output_fnames = {}
        metadata_fnames = {}
    else:
        fname_formats = get_filename_formatters(output_dict)
        output_fnames, metadata_fnames = get_output_filenames(
            fname_formats, output_dict, prod_plugin, alg_xarray, area_def
        )

    if output_plugin.family == "xarray_data":
        LOG.interactive(
            "  Producing '%s' family final outputs '%s'...",
            output_plugin.family,
            output_plugin.name,
        )
        output_products = output_plugin(
            xarray_obj=alg_xarray,
            product_names=[prod_plugin.name, "latitude", "longitude"],
            output_fnames=list(output_fnames.keys()),
            **output_kwargs,
        )
        # Disabling output_fnames check
        if output_products != list(output_fnames.keys()):
            raise ValueError("Did not produce expected products")
    else:
        mpl_colors_info = None
        if "colormapper" in prod_plugin["spec"]:
            cmap_plugin = colormappers.get_plugin(
                prod_plugin["spec"]["colormapper"]["plugin"]["name"]
            )
            cmap_args = prod_plugin["spec"]["colormapper"]["plugin"]["arguments"]
            mpl_colors_info = cmap_plugin(**cmap_args)

        output_plugin = output_formatters.get_plugin(output_formatter)
        output_kwargs = remove_unsupported_kwargs(output_plugin, output_kwargs)
        display_name = prod_plugin["spec"].get("display_name", prod_plugin.name)
        if output_plugin.family == "image":
            # This returns None if not specified
            LOG.interactive(
                "  Producing '%s' family final outputs '%s'...",
                output_plugin.family,
                output_plugin.name,
            )
            try:
                output_products = output_plugin(
                    area_def,
                    xarray_obj=alg_xarray,
                    product_name=prod_plugin.name,
                    output_fnames=list(output_fnames.keys()),
                    product_name_title=display_name,
                    mpl_colors_info=mpl_colors_info,
                    **output_kwargs,
                )
            except OutputFormatterInvalidProjectionError as e:
                LOG.warning(
                    "Failed to produce %s output: %s",
                    output_plugin.name,
                    "; ".join(list(output_fnames.keys())),
                )
                LOG.warning("OutputFormatterInvalidProjectionError: %s", e)
                output_fnames = {}
                output_products = []
            except OutputFormatterDatelineError as e:
                LOG.warning(
                    "Failed to produce %s output: %s",
                    output_plugin.name,
                    "; ".join(list(output_fnames.keys())),
                )
                LOG.warning("OutputFormatterDatelineError: %s", e)
                output_fnames = {}
                output_products = []
            if output_products != list(output_fnames.keys()):
                raise ValueError("Did not produce expected products")
        elif output_plugin.family == "unprojected":
            # This returns None if not specified
            LOG.interactive(
                "  Producing '%s' family final outputs '%s'...",
                output_plugin.family,
                output_plugin.name,
            )
            output_products = output_plugin(
                xarray_obj=alg_xarray,
                product_name=prod_plugin.name,
                output_fnames=list(output_fnames.keys()),
                product_name_title=display_name,
                mpl_colors_info=mpl_colors_info,
                **output_kwargs,
            )
            if output_products != list(output_fnames.keys()):
                raise ValueError("Did not produce expected products")
        elif output_plugin.family == "image_overlay":
            # This can include background information, feature/gridline annotations,
            # information, etc
            LOG.interactive(
                "  Producing '%s' family final outputs '%s'...",
                output_plugin.family,
                output_plugin.name,
            )
            output_products = output_plugin(
                area_def,
                xarray_obj=alg_xarray,
                product_name=prod_plugin.name,
                output_fnames=list(output_fnames.keys()),
                product_name_title=display_name,
                mpl_colors_info=mpl_colors_info,
                **output_kwargs,
            )
            if output_products != list(output_fnames.keys()):
                raise ValueError("Did not produce expected products")
        elif output_plugin.family == "xrdict_area_product_outfnames_to_outlist":
            # For xarray_dict type, pass the full fused_xarray_dict.
            output_kwargs["product_name_title"] = (display_name,)
            output_kwargs["mpl_colors_info"] = mpl_colors_info
            output_kwargs = remove_unsupported_kwargs(output_plugin, output_kwargs)
            LOG.interactive(
                "  Producing '%s' family final outputs '%s'...",
                output_plugin.family,
                output_plugin.name,
            )
            output_products = output_plugin(
                xarray_dict=fused_xarray_dict,
                area_def=area_def,
                product_name=prod_plugin.name,
                output_fnames=list(output_fnames.keys()),
                **output_kwargs,
            )
            if output_products != list(output_fnames.keys()):
                raise ValueError("Did not produce expected products")
        elif output_plugin.family == "xrdict_area_product_to_outlist":
            # For xarray_dict type, pass the full fused_xarray_dict.
            output_kwargs["product_name_title"] = (display_name,)
            output_kwargs["mpl_colors_info"] = mpl_colors_info
            output_kwargs = remove_unsupported_kwargs(output_plugin, output_kwargs)
            LOG.interactive(
                "  Producing '%s' family final outputs '%s'...",
                output_plugin.family,
                output_plugin.name,
            )
            output_products = output_plugin(
                xarray_dict=fused_xarray_dict,
                area_def=area_def,
                product_name=prod_plugin.name,
                **output_kwargs,
            )
            for output_product in output_products:
                output_fnames[output_product] = {}
            # No input filename list, no check that output filename list matches
            LOG.info(
                "Not checking output file list for output family %s",
                output_plugin.family,
            )
        else:
            raise ValueError(
                f"Unsupported output family {output_plugin.family} "
                f"for output format {output_formatter}"
            )

    all_final_products = output_all_metadata(
        output_dict, output_fnames, metadata_fnames, alg_xarray, area_def
    )

    return all_final_products


def get_area_defs_from_command_line_args(
    command_line_args, xobjs, variables=None, filter_time=True
):
    """Get area def from command line args."""
    from geoips.sector_utils.utils import (
        get_static_area_defs_for_xarray,
        get_tc_area_defs_for_xarray,
    )
    from geoips.sector_utils.utils import get_trackfile_area_defs
    from geoips.sector_utils.utils import filter_area_defs_actual_time

    sector_list = None
    tcdb_sector_list = None
    tcdb = None
    trackfile_sector_list = None
    trackfiles = None
    trackfile_parser = None
    tc_spec_template = None
    self_register_dataset = None
    self_register_source = None
    area_defs = []

    # If we are requesting an area definition that is tied directly to the reader
    # METADATA, identify it here.
    # This is useful for datasets that are pre-registered to a specific region
    # (like TCs, etc)
    if (
        "reader_defined_area_def" in command_line_args
        and command_line_args["reader_defined_area_def"]
    ):
        area_def = xobjs["METADATA"].attrs["area_definition"]

        # Provide standard area_def information that GeoIPS expects
        if not hasattr(area_def, "sector_type"):
            area_def.attrs["sector_type"] = "reader_defined"

        if not hasattr(area_def, "name"):
            setattr(area_def, "name", area_def.sector_type)

        if not hasattr(area_def, "area_id"):
            setattr(area_def, "area_id", area_def.name)

        if not hasattr(area_def, "description"):
            setattr(area_def, "description", area_def.name)

        area_defs += [area_def]
    LOG.interactive("Getting all area defs from command line args:")
    if "sector_list" in command_line_args:
        sector_list = command_line_args["sector_list"]
        LOG.interactive("  sector_list: %s", sector_list)
    if "tcdb_sector_list" in command_line_args:
        tcdb_sector_list = command_line_args["tcdb_sector_list"]
        LOG.interactive("  tcdb_sector_list: %s", tcdb_sector_list)
    if "tcdb" in command_line_args:
        tcdb = command_line_args["tcdb"]
        LOG.interactive("  tcdb: %s", tcdb)
    if "trackfile_sector_list" in command_line_args:
        trackfile_sector_list = command_line_args["trackfile_sector_list"]
        LOG.interactive("  trackfile_sector_list: %s", trackfile_sector_list)
    if "trackfiles" in command_line_args:
        trackfiles = command_line_args["trackfiles"]
        LOG.interactive("  trackfiles: %s", trackfiles)
    if "trackfile_parser" in command_line_args:
        trackfile_parser = command_line_args["trackfile_parser"]
        LOG.interactive("  trackfile_parser: %s", trackfile_parser)
    if "tc_spec_template" in command_line_args:
        tc_spec_template = command_line_args["tc_spec_template"]
        LOG.interactive("  tc_spec_template: %s", tc_spec_template)

    # This indicates that the "area_definition" will be the definition for one
    # of the native datasets
    if (
        "self_register_dataset" in command_line_args
        and "self_register_source" in command_line_args
    ):
        self_register_dataset = command_line_args["self_register_dataset"]
        self_register_source = command_line_args["self_register_source"]

    if self_register_dataset and self_register_source:
        if (
            "area_definition" in xobjs[self_register_dataset].attrs
            and xobjs[self_register_dataset].attrs["area_definition"] is not None
        ):
            area_def = xobjs[self_register_dataset].attrs["area_definition"]
        else:
            import pyresample

            orig_lons = xobjs[self_register_dataset]["longitude"]
            lons = pyresample.utils.wrap_longitudes(orig_lons)
            area_def = pyresample.geometry.SwathDefinition(
                lons=lons,
                lats=xobjs[self_register_dataset]["latitude"],
            )
            min_lat = xobjs[self_register_dataset]["latitude"].min()
            max_lat = xobjs[self_register_dataset]["latitude"].max()
            min_lon = xobjs[self_register_dataset]["longitude"].min()
            max_lon = xobjs[self_register_dataset]["longitude"].max()
            if max_lon > 180 and min_lon < 180:
                min_lon = lons.where(lons > 0).min()
                max_lon = lons.where(lons < 0).max()
            else:
                min_lon = lons.min()
                max_lon = lons.max()
            area_def.area_extent_ll = [min_lon, min_lat, max_lon, max_lat]
            if (
                "interpolation_radius_of_influence"
                in xobjs[self_register_dataset].attrs
            ):
                area_def.pixel_size_x = xobjs[self_register_dataset].attrs[
                    "interpolation_radius_of_influence"
                ]
                area_def.pixel_size_y = xobjs[self_register_dataset].attrs[
                    "interpolation_radius_of_influence"
                ]
            elif "sample_distance_km" in xobjs[self_register_dataset].attrs:
                area_def.pixel_size_x = xobjs[self_register_dataset].attrs[
                    "sample_distance_km"
                ]
                area_def.pixel_size_y = xobjs[self_register_dataset].attrs[
                    "sample_distance_km"
                ]

        if not hasattr(area_def, "sector_info"):
            setattr(
                area_def,
                "sector_info",
                {
                    "self_register_dataset": self_register_dataset,
                    "self_register_source": self_register_source,
                },
            )
        else:
            area_def.sector_info["self_register_dataset"] = self_register_dataset
            area_def.sector_info["self_register_source"] = self_register_source

        # Provide standard area_def information that GeoIPS expects
        if not hasattr(area_def, "sector_type"):
            setattr(area_def, "sector_type", "self_register")

        if not hasattr(area_def, "name"):
            setattr(area_def, "name", area_def.sector_type)

        if not hasattr(area_def, "area_id"):
            setattr(area_def, "area_id", area_def.name)

        if not hasattr(area_def, "description"):
            setattr(area_def, "description", area_def.name)

        # Add it to the list
        area_defs += [area_def]

    if sector_list:
        if xobjs is None:
            area_defs += get_static_area_defs_for_xarray(None, sector_list)
        else:
            area_defs += get_static_area_defs_for_xarray(xobjs["METADATA"], sector_list)
    if tcdb:
        if xobjs is None:
            raise (TypeError, "Must have xobjs defined for tcdb sectors")
        area_defs += get_tc_area_defs_for_xarray(
            xobjs["METADATA"],
            tcdb_sector_list=tcdb_sector_list,
            tc_spec_template=tc_spec_template,
            trackfile_parser=trackfile_parser,
            aid_type="BEST",
        )
    if trackfiles:
        area_defs += get_trackfile_area_defs(
            trackfiles,
            trackfile_parser,
            trackfile_sector_list,
            tc_spec_template,
            aid_type="BEST",
            start_datetime=xobjs["METADATA"].start_datetime - timedelta(hours=8),
            end_datetime=xobjs["METADATA"].end_datetime + timedelta(hours=3),
        )

    # If we have a "short" data file, return only a single dynamic sector closest to the
    # start time. If longer than one swath for polar orbiters, we may have more than one
    # "hit", so don't filter.
    if (
        filter_time
        and xobjs is not None
        and xobjs["METADATA"].end_datetime - xobjs["METADATA"].start_datetime
        < timedelta(hours=3)
    ):
        area_defs = filter_area_defs_actual_time(
            area_defs, xobjs["METADATA"].start_datetime
        )

    LOG.info("Allowed area_defs: %s", [ad.name for ad in area_defs])
    return list(area_defs)


def get_alg_xarray(
    sect_xarrays,
    area_def,
    prod_plugin,
    processed_xarrays=None,
    resector=True,
    resampled_read=False,
    variable_names=None,
    window_start_time=None,
    window_end_time=None,
):
    """Get alg xarray.

    Parameters
    ----------
    sect_xarrays: dict of xarray.Dataset
        dictionary of xarray Datasets to apply algorithm.
    area_def: pyresample.AreaDefinition
        Spatial region required in the final xarray Datasets.
    prod_plugin: ProductPlugin
        GeoIPS Product Plugin obtained through interfaces.products.get_plugin("name").
    resector: bool, default=True
        Specify whether to resector the data prior to applying the algorithm.
    resampled_read: bool, default=False
        Specify whether a resampled read is required, needed for datatypes that
        will be read within "get_alg_xarray"
    variable_names: list of str
        List of variable names within xarray Datasets to include in the final
        sectored xarray Datasets
    window_start_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
    window_end_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.

    Returns
    -------
    xarray.Dataset
        xarray Dataset containing the final data after interpolation, algorithm,
        resectoring, etc have been applied.
    """
    if not variable_names:
        # original input variables from sensor.py (i.e., abi.py)
        variables = get_required_variables(prod_plugin)
    else:
        # If variable_names are passed, actually use them
        # Previously was only being used for checking existence of variables in
        # sectored xarray.
        variables = variable_names

    LOG.interactive("Applying algorithms and interpolation...")

    datasets_for_vars = get_requested_datasets_for_variables(prod_plugin)

    # Get the algorithm and interpolator plugins we need to use, based
    # on the contents of prod_plugin.
    alg_plugin, alg_args, interp_plugin, interp_args = get_alg_and_interp_plugins(
        prod_plugin
    )

    # Re-sector the xarrays if requested.
    curr_sect_xarrays = resector_xarrays(
        resector, sect_xarrays, area_def, variables, window_start_time, window_end_time
    )

    LOG.info("get_alg_xarray required variables: %s", variables)
    LOG.info("get_alg_xarray requested datasets for variables: %s", datasets_for_vars)

    # If we want to run the algorithm prior to interpolation, apply the algorithm here,
    # and return either the unprojected result or interpolated result appropriately.
    # For the xarray dict option, it should probably be more general than JUST
    # unsectored xarray dict with algorithm / without area, but that was the
    # originally supported family. An issue could be created to support xarray
    # dict formats more generally.
    if (
        prod_plugin.family
        in ["algorithm_colormapper", "algorithm_interpolator_colormapper", "algorithm"]
        + PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITH_ALGORITHM_WITHOUT_AREA
    ):
        LOG.interactive("  Applying algorithms...")
        # All of these apply the algorithm first, so go ahead and
        # apply the algorithm.
        # Some use variables and some use variable_names... So pass both.
        # Some use curr_sect_xarrays and some use sect_xarrays... So pass both.
        # MLS I'm guessing this random selection is not right.
        alg_xarray = apply_alg_first(
            alg_plugin,
            alg_args,
            prod_plugin,
            curr_sect_xarrays,
            sect_xarrays,
            variables,
            variable_names,
            area_def,
        )

        if prod_plugin.family in ["algorithm_interpolator_colormapper"]:
            # Now apply the intepolator after applying the algorithm.
            final_xarray = apply_interp_after_alg(
                alg_xarray,
                interp_plugin,
                interp_args,
                prod_plugin,
                area_def,
                processed_xarrays,
            )
        else:
            final_xarray = alg_xarray

        # MLS This appears to update alg_xarray, not final_xarray, with the
        # area_def information.  So this might not be right..
        final_xarray = add_attrs_from_area_def(final_xarray, alg_xarray, area_def)
        # return here - we are done for either alg_cmap or alg_interp_cmap type
        return final_xarray

    # NOTE if algorithm specified first in product_type, we will not get to this point!
    # Returned from above if statement
    LOG.interactive("  Applying interpolators...")

    # For products that first require interpolator, start with interpolator.
    # These appear to consistently use curr_sect_xarrays and variables.
    # Don't re-interpolate if a variable is found in processed_xarrays.
    interp_xarray = apply_interp_first(
        variables,
        curr_sect_xarrays,
        prod_plugin,
        datasets_for_vars,
        resampled_read,
        area_def,
        processed_xarrays,
    )

    # Make sure we have all the appropriate attributes attached to the current
    # interp_xarray.
    # Use force=False so if attributes were set above, we do not overwrite them.
    # MLS This was originally using sect_xarray, which would have just been
    # the last sect_xarray when looping through the datasets.  Not sure that
    # is what we wanted. Also not sure how much this is going to change outputs.
    copy_standard_metadata(sect_xarrays["METADATA"], interp_xarray, force=False)

    LOG.interactive("  Applying algorithms after interpolators...")

    # Now apply the algorithm after interpolating. Don't reapply if it is
    # found in processed_xarrays.
    interp_xarray = apply_alg_after_interp(
        interp_xarray,
        area_def,
        alg_plugin,
        alg_args,
        prod_plugin,
        variables,
        processed_xarrays,
    )

    # Make sure we have all the appropriate attributes attached to the current
    # interp_xarray.
    # Use force=False so if attributes were set above, we do not overwrite them.
    # MLS This was originally using sect_xarray, which would have just been
    # the last sect_xarray when looping through the datasets.  Not sure that
    # is what we wanted. Also not sure how much this is going to change outputs.
    LOG.interactive("  Setting metadata on final xarray...")
    copy_standard_metadata(sect_xarrays["METADATA"], interp_xarray, force=False)
    # Attach final product_name to the interp_xarray as well (the end goal of
    # this routine).
    if (
        interp_xarray.attrs.get("product_name")
        and interp_xarray.attrs["product_name"] != prod_plugin.name
    ):
        # Only include product_names if there is more than one product_name
        if "product_names" not in interp_xarray.attrs:
            interp_xarray.attrs["product_names"] = [prod_plugin.name]
        else:
            interp_xarray.attrs["product_names"].append(prod_plugin.name)
    interp_xarray.attrs["product_name"] = prod_plugin.name
    # MLS This one uses interp_xarray and interp_xarray...
    # The other one I think passed the wrong ones.
    interp_xarray = add_attrs_from_area_def(interp_xarray, interp_xarray, area_def)

    return interp_xarray


def verify_area_def(
    area_defs,
    check_area_def,
    data_start_datetime,
    data_end_datetime,
    time_range_hours=3,
):
    """Verify current area definition is the closest to the actual data time.

    When looping through multiple dynamic area definitions for a full data file that
    temporally covers more than one dynamic area_def, there is no way of knowing
    which dynamic area_def has the best coverage until AFTER we have actually
    sectored the data to the specific area_def.

    Call this utility on the current area_def (check_area_def) for the sectored
    data file, plus the full list of area definitions (area_defs) that cover the
    FULL data file.

    Returns
    -------
    bool

        * True if the current area definition is NOT dynamic
        * True if the current area definition IS dynamic and
          is the closest temporally to the sectored data.
        * False if the current area definition is removed when
          filtering the list of area definitions based on the
          actual sectored data time.

    """
    retval = True
    # If this is not a dynamic sector, then there is no need to check the
    # temporal matching between the data time and the sector time, because
    # there is no sector time to check.
    if not is_dynamic_sector(check_area_def):
        retval = True
    # If the data coverage is more than time_range_hours, do not check
    # because it may be ambiguous which area definition is actually the "closest".
    elif data_end_datetime - data_start_datetime < timedelta(hours=time_range_hours):
        new_area_defs = filter_area_defs_actual_time(area_defs, data_start_datetime)
        LOG.info("Allowed area_defs: %s", [ad.name for ad in new_area_defs])
        if check_area_def.name not in [ad.name for ad in new_area_defs]:
            retval = False

    return retval


def call(fnames, command_line_args=None):
    """Workflow for running products from a single data source.

    Parameters
    ----------
    fnames : list
        List of strings specifying full paths to input file names to process
    command_line_args : dict
        dictionary of command line arguments

    Returns
    -------
    list
        Return list of strings specifying full paths to output products that
        were produced

    See Also
    --------
    ``geoips.commandline.args``
        Complete list of available command line args.
    """
    from datetime import datetime

    process_datetimes = {}
    process_datetimes["overall_start"] = datetime.utcnow()
    final_products = []
    removed_products = []
    saved_products = []
    database_writes = []
    ss_pid = getpid()

    pid_track = PidLog(ss_pid, logstr="MEMUSG")

    from geoips.commandline.args import check_command_line_args

    # These args should always be checked
    check_args = [
        "sector_list",
        "tcdb",
        "tcdb_sector_list",  # TC Database sectors,
        "trackfiles",
        "trackfile_parser",
        "trackfile_sector_list",  # Flat text trackfile,
        "reader_name",
        "product_name",
        "gridline_annotator",
        "feature_annotator",
        "product_spec_override",
        "output_formatter",
        "filename_formatter",
        "output_formatter_kwargs",
        "filename_formatter_kwargs",
        "output_checker_kwargs",
        "metadata_output_formatter",
        "metadata_filename_formatter",
        "metadata_output_formatter_kwargs",
        "metadata_filename_formatter_kwargs",
        "no_presectoring",
        "sector_adjuster",
        "sector_adjuster_kwargs",
        "reader_defined_area_def",
        "self_register_source",
        "self_register_dataset",
        "window_start_time",
        "window_end_time",
        "sectored_read",
        "resampled_read",
        "product_db",
    ]

    check_command_line_args(check_args, command_line_args)

    if not fnames:
        raise IOError(
            "No files found on disk. "
            "Please include valid files either at the command line, "
            "or within the YAML config"
        )
    for fname in fnames:
        if not exists(fname):
            raise IOError(
                f"File '{fname}' not found. "
                "Please include valid files either at the command line, "
                "or within the YAML config"
            )

    window_start_time = command_line_args.get("window_start_time")
    window_end_time = command_line_args.get("window_end_time")
    product_name = command_line_args["product_name"]  # 89HNearest
    output_formatter = command_line_args[
        "output_formatter"
    ]  # output_formatters.imagery_annotated
    reader_name = command_line_args["reader_name"]  # ssmis_binary
    reader_kwargs = command_line_args.get("reader_kwargs")
    if not reader_kwargs:
        reader_kwargs = {}
    compare_path = command_line_args["compare_path"]
    output_file_list_fname = command_line_args["output_file_list_fname"]
    sector_adjuster = command_line_args["sector_adjuster"]
    sector_adjuster_kwargs = command_line_args["sector_adjuster_kwargs"]
    self_register_source = command_line_args["self_register_source"]
    self_register_dataset = command_line_args["self_register_dataset"]
    reader_defined_area_def = command_line_args["reader_defined_area_def"]
    sectored_read = command_line_args["sectored_read"]
    resampled_read = command_line_args["resampled_read"]
    product_db = command_line_args["product_db"]
    product_db_writer = command_line_args["product_db_writer"]
    product_db_writer_kwargs = command_line_args["product_db_writer_kwargs"]
    presector_data = not command_line_args["no_presectoring"]
    output_checker_kwargs = command_line_args["output_checker_kwargs"]

    if product_db:
        from geoips.interfaces import databases

        db_writer = databases.get_plugin(product_db_writer)
        if not getenv("GEOIPS_DB_URI"):
            raise ValueError("Need to set $GEOIPS_DB_URI")

    # Load plugins
    reader_plugin = readers.get_plugin(reader_name)

    pid_track.print_mem_usg()

    num_jobs = 0
    LOG.interactive(
        "Reading metadata from dataset with reader '%s'...", reader_plugin.name
    )
    xobjs = reader_plugin(fnames, metadata_only=True, **reader_kwargs)
    source_name = xobjs["METADATA"].source_name
    pid_track.print_mem_usg()

    prod_plugin = products.get_plugin(
        source_name, product_name, command_line_args["product_spec_override"]
    )

    variables = get_required_variables(prod_plugin)

    # If we need to pull area_defs from the reader, then we need to read in
    # order to determin what to run
    if (not sectored_read and not resampled_read) and (
        reader_defined_area_def or (self_register_source and self_register_dataset)
    ):
        LOG.interactive(
            "Reading full dataset "
            "for self registered products "
            "with reader '%s'...",
            reader_plugin.name,
        )
        xobjs = reader_plugin(
            fnames, metadata_only=False, chans=variables, **reader_kwargs
        )

    # Use the xarray objects and command line args to determine required area_defs
    pid_track.print_mem_usg()
    area_defs = get_area_defs_from_command_line_args(
        command_line_args, xobjs, variables, filter_time=True
    )

    # If we do not need to pull area_defs from the reader, read the data AFTER
    # we determine we have areas to run
    if area_defs and (
        not reader_defined_area_def
        and not self_register_source
        and not sectored_read
        and not resampled_read
    ):
        pid_track.print_mem_usg()
        LOG.interactive("Reading full dataset with reader '%s'...", reader_plugin.name)
        xobjs = reader_plugin(
            fnames, metadata_only=False, chans=variables, **reader_kwargs
        )

    pid_track.print_mem_usg()
    # If we have a product of a family that does not require an area definition,
    # and operates on dictionaries of xarrays, process it here.
    # This will not have any required area_defs, so will never make it into
    # the loop through all area_defs below.
    if prod_plugin.family in PRODUCT_FAMILIES_OF_XARRAY_DICT_WITHOUT_AREA:
        LOG.interactive(
            "Reading full dataset for unsectored products with reader '%s'...",
            reader_plugin.name,
        )
        xdict = reader_plugin(fnames, metadata_only=False, **reader_kwargs)
        # If this product family requires an algorithm to be applied, apply here.
        if (
            prod_plugin.family
            in PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITH_ALGORITHM_WITHOUT_AREA
        ):
            xdict = get_alg_xarray(
                xdict,
                area_def=None,
                prod_plugin=prod_plugin,
                resector=False,
                resampled_read=resampled_read,
                window_start_time=window_start_time,
                window_end_time=window_end_time,
            )
        # This is a workaround so these products can use single source and other
        # algorithms which don't return a dictionary of xarrays. Just convert this back
        # to a dictionary under the hood and hope the metadata included in 'xdict.attrs'
        # is enough for your filename formatter.
        if not isinstance(xdict, dict):
            xdict = {prod_plugin.name: xdict, "METADATA": xdict[[]]}

        final_products += process_xarray_dict_to_output_format(
            xdict, variables, prod_plugin, command_line_args
        )
    # If we do NOT want to apply an algorithm to a dataset, but we DO want
    # to allow sectoring to a given area, just read in the data at this point,
    # and sectoring will be handled within the loop over all area defs below.
    # This looks like we are missing an option for applying an algorithm
    # and sectoring the dataset to a given area. We may need to create an
    # issue to resolve this missing "order of operations" if anyone needs it
    # before the order based procflow is finalized.
    elif (
        prod_plugin.family
        in PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITH_AREA
    ):
        LOG.interactive(
            "Reading full dataset for unsectored products requiring area "
            "definition, using reader '%s'...",
            reader_plugin.name,
        )
        # Note we only read in the data in this case, and do NOT process to
        # the final products.  Since we use the area definitions for these
        # product families, the final products will be obtained during the
        # loop through all area defs below.
        xdict = reader_plugin(fnames, metadata_only=False, **reader_kwargs)

    pid_track.print_mem_usg()

    new_attrs = {"filename_extra_fields": {}}
    # This is the main loop over all area defs - used for any processing that
    # either sectors or interpolates the datasets (even products that self
    # register to one of the existing datasets will use area definitions -
    # an area definition is created from the dataset that is being self
    # registered to in that case).  Only cases where no sectoring or
    # interpolation are needed will bypass this loop
    # (PRODUCT_FAMILIES_OF_XARRAY_DICT_WITHOUT_AREA).
    for area_def in area_defs:
        # If the current product is of family that DOES NOT require sectoring,
        # DOES NOT apply an algorithm, and DOES require an area_definition,
        # then process it here.
        if (
            prod_plugin.family
            in PRODUCT_FAMILIES_OF_UNSECTORED_XARRAY_DICT_WITHOUT_ALGORITHM_WITH_AREA
        ):
            LOG.interactive(
                "Producing outputs for unsectored product '%s'...", prod_plugin.name
            )
            final_products += process_xarray_dict_to_output_format(
                xdict, variables, prod_plugin, command_line_args, area_def
            )
            continue

        LOG.info("\n\n\n\nNEXT area definition: %s", area_def)
        pad_area_def = pad_area_definition(area_def, xobjs["METADATA"].source_name)

        # Only attempt to read within the area_def loop if we have requested
        # "sectored_read" or "resampled_read"
        if sectored_read or resampled_read:
            try:
                LOG.interactive(
                    "Reading sectored dataset "
                    "for self registered products "
                    "with reader '%s'...",
                    reader_plugin.name,
                )
                xobjs = reader_plugin(
                    fnames,
                    metadata_only=False,
                    chans=variables,
                    area_def=pad_area_def,
                    **reader_kwargs,
                )
            # geostationary satellites fail with IndexError when the area_def
            # does not intersect the data.  Just skip those.  We need a better
            # method for handling this generally, but for
            # now skip IndexErrors.
            except IndexError as resp:
                LOG.error("SKIPPING no coverage for %s, %s", area_def.name, str(resp))
                continue

        process_datetimes[area_def.area_id] = {}
        process_datetimes[area_def.area_id]["start"] = datetime.utcnow()
        # add satellite_azimuth_angle and solar_azimuth_angle into list of the variables
        # for ABI only (come from ABI reader)
        if area_def.sector_type in ["reader_defined", "self_register"]:
            LOG.interactive(
                "CONTINUE Not sectoring sector_type %s", area_def.sector_type
            )
            pad_sect_xarrays = xobjs
        else:
            if presector_data:
                LOG.interactive("Sectoring xarrays, variables %s...", variables)
                # Note window start/end time overrides hours before/after sector time
                pad_sect_xarrays = sector_xarrays(
                    xobjs,
                    pad_area_def,
                    varlist=variables,
                    hours_before_sector_time=6,
                    hours_after_sector_time=9,
                    drop=True,
                    window_start_time=window_start_time,
                    window_end_time=window_end_time,
                )
            else:
                pad_sect_xarrays = xobjs

        pid_track.print_mem_usg()
        if len(pad_sect_xarrays.keys()) == 0:
            LOG.interactive(
                "SKIPPING no sectored xarrays returned for %s", area_def.name
            )
            continue

        # Now we check to see if the current area_def is the closest one to
        # the dynamic time, if appropriate. We could end up with multiple
        # area_defs for a single dynamic sector, and we can't truly test to see
        # how close each one is to the actual data until we sector it...
        # So, check now to see if any of the area_defs in list_area_defs is
        # closer than pad_area_def
        if not verify_area_def(
            area_defs,
            pad_area_def,
            pad_sect_xarrays["METADATA"].start_datetime,
            pad_sect_xarrays["METADATA"].end_datetime,
        ):
            LOG.info(
                "SKIPPING duplicate area_def, out of time range, for %s", area_def.name
            )
            continue

        LOG.interactive("Producing sectored outputs...")
        curr_output_products = process_sectored_data_output(
            pad_sect_xarrays,
            variables,
            prod_plugin,
            command_line_args,
            area_def=area_def,
        )

        pid_track.print_mem_usg()
        # If we had a request for sectored data processing, skip the rest of the loop
        if curr_output_products:
            final_products += curr_output_products
            continue

        if sector_adjuster:
            sect_adj_plugin = sector_adjusters.get_plugin(sector_adjuster)
            # Use normal size sectored xarray when running sector_adjuster, not padded
            # Center time (mintime + (maxtime - mintime)/2) is very slightly
            # different for different size sectored arrays, so for consistency
            # if we change padding amounts, use the fully sectored array for
            # adjusting the area_def.
            if pad_sect_xarrays["METADATA"].source_name not in ["amsu-b", "mhs"]:
                if area_def.sector_type in ["reader_defined", "self_register"]:
                    LOG.interactive(
                        "CONTINUE Not sectoring sector_type %s", area_def.sector_type
                    )
                    sect_xarrays = pad_sect_xarrays
                else:
                    LOG.interactive("Sectoring padded xarrays...")
                    if presector_data:
                        # Note window start/end time overrides hours
                        # before/after sector time
                        sect_xarrays = sector_xarrays(
                            pad_sect_xarrays,
                            area_def,
                            varlist=variables,
                            hours_before_sector_time=6,
                            hours_after_sector_time=9,
                            drop=True,
                            window_start_time=window_start_time,
                            window_end_time=window_end_time,
                        )
                    else:
                        sect_xarrays = pad_sect_xarrays
                if (
                    sect_adj_plugin.family
                    == "list_xarray_list_variables_to_area_def_out_fnames"
                ):
                    LOG.interactive("Adjusting sectors with %s...", sector_adjuster)
                    area_def, adadj_fnames = sect_adj_plugin(
                        list(sect_xarrays.values()),
                        area_def,
                        variables,
                        **sector_adjuster_kwargs,
                    )
                else:
                    LOG.interactive("Adjusting sectors with %s...", sector_adjuster)
                    area_def = sect_adj_plugin(
                        list(sect_xarrays.values()),
                        area_def,
                        variables,
                        **sector_adjuster_kwargs,
                    )
            else:
                # AMSU-b specifically needs full swath width...
                if (
                    sect_adj_plugin.family
                    == "list_xarray_list_variables_to_area_def_out_fnames"
                ):
                    LOG.interactive("Adjusting sectors with %s...", sector_adjuster)
                    area_def, adadj_fnames = sect_adj_plugin(
                        list(pad_sect_xarrays.values()),
                        area_def,
                        variables,
                        **sector_adjuster_kwargs,
                    )
                else:
                    LOG.interactive("Adjusting sectors with %s...", sector_adjuster)
                    area_def = sect_adj_plugin(
                        list(pad_sect_xarrays.values()),
                        area_def,
                        variables,
                        **sector_adjuster_kwargs,
                    )
            # These will be added to the alg_xarray
            # new_attrs['area_definition'] = area_def
            if "adjustment_id" in area_def.sector_info:
                new_attrs["filename_extra_fields"]["adjustment_id"] = (
                    area_def.sector_info["adjustment_id"]
                )

        pid_track.print_mem_usg()
        all_vars = []
        for key, xobj in pad_sect_xarrays.items():
            # Double check the xarray object actually contains data
            for var in list(xobj.variables.keys()):
                if xobj[var].count() > 0:
                    all_vars.append(var)
        # If the required variables are not contained within the xarray objects, do not
        # attempt to process (variables in product algorithm are not available)
        if set(variables).issubset(all_vars):
            # We want to write out the padded xarray for "xarray_data" output types
            # Otherwise, we need the fully sectored output
            output_plugin = output_formatters.get_plugin(output_formatter)
            if output_plugin.family == "xarray_data":
                alg_xarray = get_alg_xarray(
                    pad_sect_xarrays,
                    pad_area_def,
                    prod_plugin,
                    resector=False,
                    resampled_read=resampled_read,
                    window_start_time=window_start_time,
                    window_end_time=window_end_time,
                )
            elif area_def.sector_type in ["reader_defined", "self_register"]:
                alg_xarray = get_alg_xarray(
                    pad_sect_xarrays,
                    pad_area_def,
                    prod_plugin,
                    resector=False,
                    resampled_read=resampled_read,
                    variable_names=variables,
                    window_start_time=window_start_time,
                    window_end_time=window_end_time,
                )
            else:
                alg_xarray = get_alg_xarray(
                    pad_sect_xarrays,
                    area_def,
                    prod_plugin,
                    resector=presector_data,
                    resampled_read=resampled_read,
                    window_start_time=window_start_time,
                    window_end_time=window_end_time,
                )

            pid_track.print_mem_usg()

            # This defaults to "covg_func" and "covg_args" - if
            # image_production_covg_* exist, it will use those.
            covg_plugin = get_covg_from_product(
                prod_plugin,
                covg_field="image_production_coverage_checker",
            )
            covg_args = get_covg_args_from_product(
                prod_plugin,
                covg_field="image_production_coverage_checker",
            )
            # Set variable_name to prod_plugin.name if not defined.
            # Always use variable_name if it is defined.
            # Remove from args so there is not a duplicate are when passing
            # (since we are passing covg_varname explicitly).
            # Note get_covg_args_from_product was updated to return a copy of
            # covg_args, so this does not impact future uses of the product.
            covg_varname = covg_args.pop("variable_name", prod_plugin.name)
            # Note variables can be specified as DATASET:VARIABLE, since this is a
            # preprocessed alg_xarray, and not a dictionary of datasets, just use the
            # variable name (we expect the correct variable will exist in this final
            # processed array)
            if ":" in covg_varname:
                covg_varname = covg_varname.split(":")[1]

            covg = covg_plugin(
                alg_xarray,
                covg_varname,
                area_def,
                **covg_args,
            )

            fname_covg_plugin = get_covg_from_product(
                prod_plugin,
                covg_field="filename_coverage_checker",
            )
            fname_covg_args = get_covg_args_from_product(
                prod_plugin,
                covg_field="filename_coverage_checker",
            )
            # Set variable_name to prod_plugin.name if not defined.
            # Always use variable_name if it is defined.
            # Remove from args so there is not a duplicate are when passing
            # (since we are passing covg_varname explicitly).
            # Note get_covg_args_from_product was updated to return a copy of
            # covg_args, so this does not impact future uses of the product.
            fname_covg_varname = fname_covg_args.pop("variable_name", prod_plugin.name)
            # Note variables can be specified as DATASET:VARIABLE, since this is a
            # preprocessed alg_xarray, and not a dictionary of datasets, just use the
            # variable name (we expect the correct variable will exist in this final
            # processed array)
            if ":" in fname_covg_varname:
                fname_covg_varname = fname_covg_varname.split(":")[1]
            fname_covg = fname_covg_plugin(
                alg_xarray,
                fname_covg_varname,
                area_def,
                **fname_covg_args,
            )

            for attrname in new_attrs:
                LOG.info(
                    "ADDING attribute %s %s to alg_xarray",
                    attrname,
                    new_attrs[attrname],
                )
                alg_xarray.attrs[attrname] = new_attrs[attrname]

            # Apply a new coverage scheme (coverage within 300km radical range
            # from TC center) to be done  ????

            minimum_coverage = 10
            command_line_minimum_coverage = get_minimum_coverage(
                prod_plugin, command_line_args
            )
            if hasattr(alg_xarray, "minimum_coverage"):
                minimum_coverage = alg_xarray.minimum_coverage
            if command_line_minimum_coverage is not None:
                minimum_coverage = command_line_minimum_coverage
            LOG.interactive(
                "Required coverage %s for product %s, actual coverage %s",
                minimum_coverage,
                prod_plugin.name,
                covg,
            )
            if covg < minimum_coverage and fname_covg < minimum_coverage:
                LOG.interactive(
                    "SKIPPING: Insufficient coverage %s / %s for "
                    "data products for %s, %s required",
                    covg,
                    fname_covg,
                    area_def.name,
                    minimum_coverage,
                )
                continue

            output_formatter_kwargs = get_output_formatter_kwargs(
                command_line_args, xarray_obj=alg_xarray, area_def=area_def
            )

            curr_products = plot_data(
                command_line_args,
                alg_xarray,
                area_def,
                prod_plugin,
                output_formatter_kwargs,
            )
            if not curr_products:
                continue

            pid_track.print_mem_usg()
            final_products += curr_products
            curr_removed_products, curr_saved_products = remove_duplicates(
                curr_products, remove_files=True
            )
            removed_products += curr_removed_products
            saved_products += curr_saved_products

            if product_db:
                for fprod, fname_fmt in curr_products.items():
                    additional_attrs = {
                        "coverage": covg,
                        "product": product_name,
                        "fileType": fprod.split(".")[-1],
                    }
                    if db_writer.family == "xarray_area_def_to_table":
                        product_added = db_writer(
                            product_filename=fprod,
                            xarray_obj=alg_xarray,
                            area_def=area_def,
                            additional_attrs=additional_attrs,
                            **product_db_writer_kwargs,
                        )
                        database_writes += [product_added]
                    elif db_writer.family == "xarray_dict_to_table":
                        product_added = db_writer(
                            product_filename=fprod,
                            xarray_dict=alg_xarray,
                            area_def=area_def,
                            additional_attrs=additional_attrs,
                            **product_db_writer_kwargs,
                        )
                        database_writes += [product_added]
                    else:
                        LOG.error(
                            "FAILED DB WRITE: Only "
                            "xarray_area_def_to_table and xarray_dict_to_table "
                            "db_writer families supported. Either reformat "
                            "db_writer plugin as correct family, or add support "
                            "for additional families in the procflow and "
                            "databases interface."
                        )

            process_datetimes[area_def.area_id]["end"] = datetime.utcnow()
            num_jobs += 1
        else:
            LOG.interactive(
                'SKIPPING No coverage or required variables "%s" for %s %s',
                variables,
                xobjs["METADATA"].source_name,
                area_def.name,
            )

    LOG.interactive(
        "\n\n\nProcessing complete! Checking outputs...\n\n",
    )
    process_datetimes["overall_end"] = datetime.utcnow()

    if output_file_list_fname:
        LOG.info("Writing successful outputs to %s", output_file_list_fname)
        with open(output_file_list_fname, "w", encoding="utf8") as fobj:
            fobj.writelines(
                "\n".join([replace_geoips_paths(fname) for fname in final_products])
            )
            # If we don't write out the last newline, then wc won't return the
            # appropriate number, and we won't get to the last file when attempting to
            # loop through
            fobj.writelines(["\n"])

    retval = 0
    if compare_path:
        from geoips.interfaces.module_based.output_checkers import output_checkers

        for output_product in final_products:
            plugin_name = output_checkers.identify_checker(output_product)
            output_checker = output_checkers.get_plugin(plugin_name)
            kwargs = {}
            if output_checker.name in output_checker_kwargs:
                kwargs = output_checker_kwargs[output_checker.name]
            retval += output_checker(
                output_checker,
                compare_path.replace("<product>", product_name)
                .replace("<procflow>", "single_source")
                .replace("<output>", output_formatter),
                [output_product],
                **kwargs,
            )

    LOG.interactive(
        "\n\n\nThe following products were produced from procflow %s\n\n",
        basename(__file__),
    )
    for output_product in final_products:
        LOG.interactive(
            "    \u001b[34mSINGLESOURCESUCCESS\033[0m %s",
            replace_geoips_paths(output_product, curly_braces=True),
        )
        if output_product in database_writes:
            LOG.interactive("    DATABASESUCCESS %s", output_product)
    for removed_product in removed_products:
        LOG.interactive("    DELETEDPRODUCT %s", removed_product)

    pid_track.print_mem_usg(verbose=True)

    LOG.interactive("READER_NAME: %s", reader_name)
    LOG.interactive("PRODUCT_NAME: %s", product_name)
    LOG.interactive("NUM_PRODUCTS: %s", len(final_products))
    LOG.interactive("NUM_DELETED_PRODUCTS: %s", len(removed_products))
    output_process_times(process_datetimes, num_jobs, job_str="single_source procflow")
    pid_track.save_exit()
    return retval



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Interpolation interface will be deprecated v2.0.

Wrapper functions for geoips product specifications.

This functionality will be replaced with a class-based implementation v2.0,
and deprecated at that time.
"""
import logging
from geoips.interfaces import colormappers, coverage_checkers

LOG = logging.getLogger(__name__)


def get_required_variables(prod_plugin):
    """Interface will be deprecated v2.0.

    Return required variables names for the input product plugin. If variables are
    combined with their dataset name, the dataset name will be stripped and only the
    variable names will be returned.

    Parameters
    ----------
    prod_plugin : ProductPlugin
        An instance of the GeoIPS ProductPlugin class.

    Returns
    -------
    required_variables : list or dict
        * If list: List of strings specifying required variables.
        * If dict: Dictionary of variable types of lists of variable names

            * {'<variable_type>': ['var1', 'var2', ... , 'varn']}
    """
    # This can either be a list or dictionary, dependent on YAML config specification
    variables = prod_plugin["spec"].get("variables")
    if variables is None:
        return []

    # Support categorizing variables in a dictionary
    if isinstance(variables, dict):
        return_vars = {}
        for vartype in variables:
            return_vars[vartype] = []
            for varname in variables[vartype]:
                if ":" in varname:
                    return_vars[vartype] += [varname.split(":")[1]]
                else:
                    return_vars[vartype] += [varname]
    # Otherwise, just a single variable
    else:
        return_vars = []
        for varname in variables:
            if ":" in varname:
                return_vars += [varname.split(":")[1]]
            else:
                return_vars += [varname]
    return return_vars


def get_requested_datasets_for_variables(prod_plugin):
    """Interface will be deprecated v2.0.

    Retrieve required datasets if specified for product variables,
    based on requested product and source

    Within product_inputs YAML specifications, variables can be
    requested with ``<DATASET>:<VARNAME>`` if
    you need a particular variable from a specific dataset.

    If ``<DATASET>:`` is not specified, the first
    variable found when looping through the datasets is used.

    Parameters
    ----------
    product_name : str
        Name of requested product (ie, 'IR-BD', '89H', 'color89Nearest', etc)
    source_name : str
        Name of requested source (ie, 'ahi', 'modis', etc)

    Returns
    -------
    datasets_for_variable : dict
        * Dictionary of

            * ``{'<variable_name>': '<requested_dataset>'}`` OR
            * ``{'variable_type': {'<variable_name>': '<requested_dataset>'}``
    """
    variables = prod_plugin["spec"]["variables"]
    return_dict = {}
    # Support categorizing variables in a dictionary - var_dict[var_type] =
    # ['var1', 'var2', 'varn']
    if isinstance(variables, dict):
        return_dict = {}
        for vartype in variables:
            return_dict[vartype] = {}
            for varname in variables[vartype]:
                dataset, variable = varname.split(":")
                if varname not in return_dict:
                    return_dict[vartype][variable] = [dataset]
                else:
                    return_dict[vartype][variable] += [dataset]
    # Otherwise just a list
    else:
        for varname in variables:
            if ":" in varname:
                dataset, variable = varname.split(":")
                if varname not in return_dict:
                    return_dict[variable] = [dataset]
                else:
                    return_dict[variable] += [dataset]
    return return_dict


def get_data_range(prod_plugin, output_dict=None):
    """Interface will be deprecated v2.0.

    Retrieve required data range for requested product

    Parameters
    ----------
    product_name : str
        Name of requested product (ie, 'IR-BD', '89H', 'color89Nearest', etc)

    Returns
    -------
    data_range : list
        List of float specifying min and max value for the output product
        ``<geoips_package>.algorithms.<algorithm_name>.alg_params['output_data_range']``
    """
    alg_args = prod_plugin["spec"]["algorithm"]["plugin"]["arguments"]
    if "output_data_range" not in alg_args:
        raise TypeError(
            f"Product {prod_plugin.name} does not define 'output_data_range' for its "
            f"algorithm."
        )
    # Add .copy() so we aren't returning the ACTUAL list attached to the args
    return alg_args["output_data_range"].copy()


def get_cmap_data_range(prod_plugin, output_dict=None):
    """Interface will be deprecated v2.0.

    Retrieve required data range for requested product

    Parameters
    ----------
    product_name : str
        Name of requested product (ie, 'IR-BD', '89H', 'color89Nearest', etc)

    Returns
    -------
    data_range : list
        List of float specifying min and max value for the output product
        ``<geoips_package>.colormapper.<cmap_name>.arguments['data_range']``
    """
    cmap_args = prod_plugin["spec"]["colormapper"]["plugin"]["arguments"]
    if "data_range" not in cmap_args:
        raise TypeError(
            f"Product {prod_plugin.name} does not define 'output_data_range' for its "
            f"algorithm."
        )
    # Add .copy() so we aren't returning the ACTUAL list attached to the args
    return cmap_args["data_range"].copy()


def get_product_display_name(prod_plugin, output_dict=None):
    """Interface will be deprecated v2.0.

    Retrieve product display name. For titles, etc.

    Parameters
    ----------
    product_name : str
        Name of requested product (ie, 'IR-BD', '89H', 'color89Nearest', etc)
    source_name : str
        Name of requested source (ie, 'ahi', 'modis', etc)

    Returns
    -------
    product_display_name : str
        Return display name for given product

    See Also
    --------
    ``geoips.dev.check_cmap_func``
        additional information on colormapper types, arguments, and return values
    """
    try:
        return prod_plugin["spec"]["display_name"]
    except KeyError:
        return prod_plugin.name


def get_cmap_from_product(prod_plugin, output_dict=None):
    r"""Interface will be deprecated v2.0.

    Retrieve colormapper information, based on requested product and source

    Parameters
    ----------
    product_name : str
        Name of requested product (ie, 'IR-BD', '89H', 'color89Nearest', etc)
    source_name : str
        Name of requested source (ie, 'ahi', 'modis', etc)

    Returns
    -------
    cmap_func(\*\*cmap_args) : function
        Return actual colormapper information

    See Also
    --------
    ``geoips.dev.check_cmap_func``
        additional information on colormapper types, arguments, and return values
    """
    cmap_func = colormappers.get_plugin(
        prod_plugin["spec"]["colormapper"]["plugin"]["name"]
    )
    cmap_args = prod_plugin["spec"]["colormapper"]["plugin"]["arguments"]

    return cmap_func(**cmap_args)


def get_covg_from_product(prod_plugin, output_dict=None, covg_field=None):
    """Interface will be deprecated v2.0.

    Retrieve coverage check function name, based on requested product and source

    Parameters
    ----------
    product_name : str
        Name of requested product (ie, 'IR-BD', '89H', 'color89Nearest', etc)
    source_name : str
        Name of requested source (ie, 'ahi', 'modis', etc)

    Returns
    -------
    covg_func_name : str
        Return name of coverage function required for given product/source

    See Also
    --------
    ``geoips.dev.check_cmap_func``
        additional information on colormapper types, arguments, and return values
    """
    prod_spec = prod_plugin["spec"]
    if covg_field in prod_spec:
        return coverage_checkers.get_plugin(prod_spec[covg_field]["plugin"]["name"])
    elif "coverage_checker" in prod_spec:
        return coverage_checkers.get_plugin(
            prod_spec["coverage_checker"]["plugin"]["name"]
        )
    else:
        return coverage_checkers.get_plugin("masked_arrays")


def get_covg_args_from_product(prod_plugin, output_dict=None, covg_field=None):
    """Interface will be deprecated v2.0.

    Retrieve coverage check function args, based on requested product and source

    Parameters
    ----------
    product_name : str
        Name of requested product (ie, 'IR-BD', '89H', 'color89Nearest', etc)
    source_name : str
        Name of requested source (ie, 'ahi', 'modis', etc)

    Returns
    -------
    covg_func_name : str
        Return dictionary of coverage args required for given product/source

    See Also
    --------
    ``geoips.dev.check_cmap_func``
        additional information on colormapper types, arguments, and return values
    """
    prod_spec = prod_plugin["spec"]
    # Add .copy() to coverage args so we aren't passing the ACTUAL coverage
    # args attached to the product plugin.
    if covg_field in prod_spec:
        return prod_spec[covg_field]["plugin"]["arguments"].copy()
    elif "coverage_checker" in prod_spec:
        return prod_spec["coverage_checker"]["plugin"]["arguments"].copy()
    else:
        return {}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read EWS-G data.

This EWS-G(Electro-Optical Infrared Weather System - Geostationary) reader is
designed for reading theEWS-G data files (EWS-G is renamed from GOES-13).
The reader is only using the python functions and xarray variables.
The reader is based on EWS-G data in netcdf4 format.

 V1.0:  NRL-Monterey, 02/25/2021

EWS-G file information::

    Example of the gvar filename:   2020.1212.0012.goes-13.gvar.nc

    Note that channel-3 is not available for EWS-G.
      gvar_Ch3(TIR=5.8-7.3um, ctr=6.48um,4km): unit=temp-deg(C), scale_factor=0.01

    variables:
      gvar_Ch1(VIS=0.55-0.75um, ctr=0.65um,1km): unit=albedo*100,  scale_factor=0.01
      gvar_Ch2(MWIR=3.8-4.0um,  ctr=3.9um, 4km): unit=temp-deg(C), scale_factor=0.01
      gvar_Ch4(TIR=10.2-11.2um, ctr=10.7um,4km): unit=temp-deg(C), scale_factor=0.01
      gvar_Ch6(TIR=12.9-13.7um, ctr=13.3um 4km): unit=temp-deg(C), scale_factor=0.01
      latitude: unit=degree
      longitude:unit=degree
      sat_zenith: unit=degree
      sun_zenith: unit=degree
      rel_azimuth:unit=degree

      variable array definition:  var(scan,pix); scan-->lines, pix-->samples

    attributes: many
"""
# Python Standard Libraries
import logging
import os

# Third-Party Libraries
import calendar
import numpy as np
import pandas as pd
import xarray as xr

from geoips.interfaces import readers
from geoips.utils.context_managers import import_optional_dependencies

# If this reader is not installed on the system, don't fail altogether, just skip this
# import. This reader will not work if the import fails, and the package will have to be
# installed to process data of this type.

LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import netCDF4 as ncdf


# @staticmethod                                     # not sure where it is was used?

# gvar_ch6 has only half scanlines of other channels (1,2,4), we temporally do not read
# the ch6 in. We will modify this reader if gvar_ch6 is needed in the future.
VARLIST = [
    "gvar_ch1",
    "gvar_ch2",
    "gvar_ch4",
    "latitude",
    "longitude",
    "sun_zenith",
    "sat_zenith",
    "rel_azimuth",
]

# setup needed to convert var_name used in geoips: i.e., solar_zenith_angle (not
# sun_zenith) is used.
xvarnames = {
    "sun_zenith": "solar_zenith_angle",
    "sat_zenith": "satellite_zenith_angle",
    "rel_azimuth": "satellite_azimuth_angle",
}

interface = "readers"
family = "standard"
name = "ewsg_netcdf"
source_names = ["gvar"]


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read EWS-G data in netcdf4 format.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    return readers.read_data_to_xarray_dict(
        fnames,
        call_single_time,
        metadata_only,
        chans,
        area_def,
        self_register,
    )


def call_single_time(
    fnames, metadata_only=False, chans=None, area_def=None, self_register=False
):
    """Read EWS-G data in netcdf4 format for one or more files.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # --------------- loop input files ---------------
    xarray_ewsg = xr.Dataset()
    xarray_ewsg.attrs["source_file_names"] = []
    # LOG.info('Requested Channels: %s', chans)

    for fname in fnames:
        # check for a correct goes-13 data file
        data_name = os.path.basename(fname).split("_")[-1].split(".")[-1]
        if data_name != "nc":
            LOG.info("Warning: EWS-G data type:  data_type=", data_name)
            raise

        # open the paired input files
        ncdf_file = ncdf.Dataset(str(fname), "r")
        LOG.info("    Trying file %s", fname)

        if ncdf_file.satellite in ["goes-13", "goes-14", "goes-15"]:
            LOG.info("found a NOAA EWS-G data file")
        else:
            LOG.info("not a NOAA EWS-G data file: skip it")
            raise ValueError(
                f"Found {ncdf_file.satellite}, "
                "expected either goes-13, goes-14, or goes-15"
            )

        # setup attributes
        # use fname to get an initial info of  year, month, day
        # test hour/minute/second of "start_datetime" info from start_time (second of
        # the day). add scan_time info to determine the "end_datetime". scan_time has
        # time of scans for this input data file(one obs).  If end_time of this data >24
        # hr, modify value of "day" from fname.  Note: scan_time units are seconds from
        # (start_time + time_adjust)

        # date info from fname
        #        1  2  3  4  5  6  7  8  9  10 11 12
        days_mo = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        days_mo_lp = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]  # Leap year

        # date information is not contained in the data so we have to get it from
        # filename
        data_name = os.path.basename(fname)
        yr = int(data_name.split(".")[0])
        mo = int(data_name.split(".")[1][0:2])
        dy = int(data_name.split(".")[1][2:4])
        # hr = int(data_name.split(".")[2][0:2])
        # mm = int(data_name.split(".")[2][2:4])

        # determine a Leap Year?
        if calendar.isleap(yr):
            days = days_mo_lp[mo - 1]
        else:
            days = days_mo[mo - 1]

        # second of the date for this file
        start_time = (
            ncdf_file.start_time + ncdf_file.time_adjust + ncdf_file.scan_time[0]
        )
        end_time = (
            ncdf_file.start_time + ncdf_file.time_adjust + ncdf_file.scan_time[29]
        )

        yr_s = yr
        yr_e = yr
        mo_s = mo
        mo_e = mo
        dy_s = dy
        dy_e = dy
        hr_s = int(start_time / 3600)
        mm_s = int((start_time - hr_s * 3600) / 60)
        ss_s = int(start_time - (hr_s * 3600 + mm_s * 60))
        hr_e = int(end_time / 3600)
        mm_e = int((end_time - hr_e * 3600) / 60)
        ss_e = int(end_time - (hr_e * 3600 + mm_e * 60))

        if hr_s >= 24:
            # dy_s_ = dy + 1  # forward to the next date
            hr_s = hr_s - 24
            if dy_s > days:  # move to next mon
                mo_s = mo + 1
                if mo_s > 12:  # move to near year
                    yr_s = yr + 1
        if hr_e >= 24:
            # dy_e_ = dy + 1  # forward to the next date
            hr_e = hr_e - 24
            if dy_e > days:  # move to next mon
                mo_e = mo + 1
                if mo_e > 12:  # move to near year
                    yr_e = yr + 1

        start_scan = "%04d%02d%02d%02d%02d%02d" % (yr_s, mo_s, dy_s, hr_s, mm_s, ss_s)
        end_scan = "%04d%02d%02d%02d%02d%02d" % (yr_e, mo_e, dy_e, hr_e, mm_e, ss_e)

        # convert date in required format
        Start_date = pd.to_datetime(start_scan, format="%Y%m%d%H%M%S")
        End_date = pd.to_datetime(end_scan, format="%Y%m%d%H%M%S")

        xarray_ewsg.attrs["start_datetime"] = Start_date
        xarray_ewsg.attrs["end_datetime"] = End_date
        # Goes VARiable (gvar) data is on GOES-12/13 - older GOES satellites were gvissr
        # ncdf_file.sensor_name is gvissr - I think that is a mistake on
        # Terascan's part - hold out from years ago
        xarray_ewsg.attrs["source_name"] = "gvar"

        if ncdf_file.satellite == "goes-13":
            xarray_ewsg.attrs["platform_name"] = "ews-g1"
        if ncdf_file.satellite == "goes-15":
            xarray_ewsg.attrs["platform_name"] = "ews-g2"
        if ncdf_file.satellite == "goes-14":
            xarray_ewsg.attrs["platform_name"] = "ews-g3"
        xarray_ewsg.attrs["legacy_platform_name"] = ncdf_file.satellite
        xarray_ewsg.attrs["data_provider"] = "noaa"
        xarray_ewsg.attrs["source_file_names"] = [
            os.path.basename(fname) for fname in fnames
        ]

        # MTIFs need to be "prettier" for PMW products, so 2km resolution for
        # final image
        xarray_ewsg.attrs["sample_distance_km"] = 2
        xarray_ewsg.attrs["interpolation_radius_of_influence"] = 3000
        # Just return the metadata
        if metadata_only:
            # close the files
            ncdf_file.close()
            return {"METADATA": xarray_ewsg}
        # Otherwise, get the actual data, as well as the metadata
        # *************** input VIRRS variables  and output xarray required by geo

        # for varname in ncdf_file.variables.keys():
        for var in VARLIST:
            varname = var
            data = ncdf_file[varname]
            # masked_data=np.ma.masked_equal(
            #   ncdf_file[varname],ncdf_file[varname].missing_value
            # )
            masked_data = np.ma.masked_equal(data, data.missing_value)

            if var in xvarnames:
                varname = xvarnames[var]  # rename zenith/azimuth-related variables

            xarray_ewsg[varname] = xr.DataArray(masked_data)
            """
            # scale_factor should not be applied
            if 'scale_factor' in data.ncattrs():
                # apply scale_factor correction
                #xarray_ewsg[varname]=xarray_ewsg[varname]*ncdf_file[varname].scale_factor  # NOQA
                xarray_ewsg[varname]=xarray_ewsg[varname]*data.scale_factor
            """
            # convert unit from degree to Kelvin for ch2, ch4, and ch6 (ch1 is in unit
            # of albedo)
            if varname in ["gvar_ch2", "gvar_ch4", "gvar_ch6"]:
                xarray_ewsg[varname] = xarray_ewsg[varname] + 273.15
                xarray_ewsg[varname].attrs["units"] = "Kelvin"

        # close the files
        ncdf_file.close()

    return {"LOW": xarray_ewsg, "METADATA": xarray_ewsg[[]]}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Builds geoips and geoips plugin documentation."""

import tempfile
import importlib.resources
import importlib.util
import warnings
import logging
import logging.handlers
import argparse
import shutil
import os
from subprocess import call

import brassy.actions.build_release_notes as brassy_build
import brassy.utils.CLI  # noqa # because of a brassy bug; will be fixed in next vers
from rich.logging import RichHandler
from rich.traceback import install as install_rich_tracebacks
from rich.logging import Console
from rich.progress import Progress
import rich_argparse
import pygit2
import yaml
import sphinx.cmd.build as sphinx_build_module
from sphinx.ext.apidoc import main as sphinx_apidoc

from update_release_note_index import main as generate_release_note_index

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def init_logger(use_rich):
    """
    Initialize and configure the logger.

    Parameters
    ----------
    use_rich : bool
        If True, sets up rich logging else use standard stream logging

    Returns
    -------
    logger : logging.Logger
        The configured logger instance
    """
    if use_rich:
        install_rich_tracebacks()
        logging_handlers = [RichHandler(rich_tracebacks=True)]
    else:
        logging_handlers = [logging.StreamHandler()]

    logging.basicConfig(level=logging.DEBUG, datefmt="[%X]", handlers=logging_handlers)
    logger = logging.getLogger("build_docs")
    return logger


def parse_args_with_argparse():
    """
    Parse and validate command-line arguments provided to the script.

    Returns
    -------
    args : argparse.Namespace
        An object containing the parsed command-line arguments.
    """
    # Initialize parser with an example usage in the description
    parser = argparse.ArgumentParser(
        description=(
            "Build Sphinx documentation for GeoIPS related packages.\n"
            "Example: build_docs.py /path/to/repo geoips"
        ),
        formatter_class=rich_argparse.RichHelpFormatter,
    )

    # Required positional arguments
    parser.add_argument(
        "package_name",
        type=str,
        help="Name of the package to build (e.g., geoips, data_fusion).",
    )

    # Optional argument: path to documentation templates
    parser.add_argument(
        "--geoips-docs-path",
        type=str,
        default=None,
        help="Path to GeoIPS documentation templates.",
    )
    parser.add_argument(
        "--license-url",
        type=str,
        default=None,
        help=(
            "Path to GeoIPS license."
            "Default is https://github.com/NRLMMD-GEOIPS/[package_name]"
        ),
    )

    parser.add_argument(
        "--output-dir",
        type=str,
        default=None,
        help="Output dir to write built docs to",
    )
    parser.add_argument(
        "--save-temp-dir",
        type=str,
        default=None,
        help="Output dir to write temporary build files to (for debugging purposes)",
    )
    parser.add_argument(
        "--repo-path",
        type=str,
        default=None,
        help="Path to the repository (e.g., /path/to/geoips)."
        "If not provided uses [package_name] to find the packages directory"
        "Must be an existing directory.",
    )

    parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Replace output dir if it already exists",
    )

    # Parse arguments
    args = parser.parse_args()

    if not args.repo_path:
        try:
            package_path = os.path.dirname(
                str(importlib.resources.files(args.package_name))
            )
            if not package_path:
                raise ModuleNotFoundError
            pygit2.Repository(package_path)
            args.repo_path = package_path
        except ModuleNotFoundError as e:
            raise e(f"Could not automatically find repo_path for {args.package_name}.")
        except pygit2.GitError:
            raise pygit2.GitError(
                "Could not automatically find usable repo_path for "
                f"{args.package_name}. Found {package_path} but it is not a git repo"
            )

    if not args.output_dir:
        output_dir = os.getenv("GEOIPS_DOCSDIR", None)
        if output_dir:
            warnings.warn(
                f"Using output dir value {output_dir} from environmental variable "
                "$GEOIPS_DOCSDIR. This functionality is DEPRECATED and will be "
                "removed. Please pass $GEOIPS_DOCSDIR as --output-dir $GEOIPS_DOCSDIR "
                "for the same functionality."
            )
        else:
            output_dir = os.path.join(args.repo_path, "build", "sphinx", "html")
        args.output_dir = output_dir

    if args.package_name == "geoips":
        args.geoips_docs_path = os.path.join(args.repo_path, "docs")
    elif not args.geoips_docs_path:
        docs_path_env = (
            os.getenv("GEOIPS_PACKAGES_DIR", "") + "/geoips/docs"
            if os.getenv("GEOIPS_PACKAGES_DIR")
            else None
        )
        if docs_path_env:
            warnings.warn(
                "GeoIPS docs path set to $GEOIPS_PACKAGES_DIR/geoips/docs, but this "
                "fall back is deprecated. Please start passing with "
                "--geoips-docs-path $GEOIPS_PACKAGES_DIR"
            )
            args.geoips_docs_path = docs_path_env
        else:
            raise argparse.ArgumentError(
                "Please pass geoips docs path to " "build plugin documentation"
            )

    if not args.license_url:
        repo_url = os.getenv("GEOIPS_REPO_URL")
        if repo_url:
            warnings.warn(
                "Using the environmental variable $GEOIPS_REPO_URL as value for "
                "--license-url. This functionality is deprecated and will be removed. "
                "Please pass in using --license-url $GEOIPS_REPO_URL. Alternatively,"
                "you can unset $GEOIPS_REPO_URL and the URL will default to"
                "https://github.com/NRLMDD-GEOIPS/package_name"
            )
        else:
            repo_url = "https://github.com/NRLMMD-GEOIPS/" + args.package_name
        args.license_url = repo_url

    return args


def validate_path_exists_and_is_git_repo(repo_path, logger=logging.getLogger(__name__)):
    """
    Ensure given path exists and is a valid Git repository.

    Parameters
    ----------
    repo_path : str
        The file system path to validate.
    logger : logging.Logger, optional
        Logger for logging messages; defaults to the module logger.

    Raises
    ------
    FileNotFoundError
        If the provided path does not exist.
    pygit2.GitError
        If the path is not a valid Git repository.
    """
    if not os.path.exists(repo_path):
        logger.critical(f"Path does not exist: {repo_path}")
        raise FileNotFoundError
    try:
        pygit2.Repository(repo_path)
    except pygit2.GitError as e:
        logger.critical(f"{repo_path} is not a valid git repo")
        raise e


def validate_package_is_installed(package_name, logger=logging.getLogger(__name__)):
    """
    Verify that the specified package is installed in the current Python environment.

    Parameters
    ----------
    package_name : str
        The name of the package to check for installation.
    logger : logging.Logger, optional
        Logger for logging messages; defaults to the module logger.

    Raises
    ------
    ModuleNotFoundError
        If the package is not found in the current Python environment.
    """
    if importlib.util.find_spec(package_name) is None:
        logger.critical(f"ERROR: Package {package_name} is not installed")
        raise ModuleNotFoundError


def get_section_replace_string(section):
    """
    Create a placeholder string used in the index template for a given section.

    Parameters
    ----------
    section : str
        The name of the section.

    Returns
    -------
    str
        A placeholder string corresponding to the section.

    """
    return section.upper() + "_OPTIONAL"


def get_sections(package_name):
    """
    Return sections.

    Returns a list of required and optional sections.

    Returns
    -------
    required_sections : list of str
        List of required section names.
    optional_sections : list of str
        List of optional section names.

    Notes
    -----
    This could potentially be from a config file in the future. For now, it is hard
    coded like the original build_docs.sh script.
    """
    with open(os.path.join(__location__, "docs-sections.yaml"), "r") as f:
        section_data = yaml.safe_load(f)

    section_data["optional"] = section_data["current"] + section_data["legacy"]

    return (
        (x.replace("PKGNAME", package_name) for x in section_data[category])
        for category in ["required", "optional"]
    )


def return_file_content(template_path):
    """
    Render file into a string.

    This function reads the content of the template file
    and returns the resulting string.

    Parameters
    ----------
    template_path : str
        The file path to the template.

    Returns
    -------
    str
        The rendered template content as a string.
    """
    with open(template_path, "rt") as template_file:
        return template_file.read()


def update_content_for_section(
    build_dir, content, section, is_optional=False, log=logging.getLogger(__name__)
):
    """
    Replace placeholders in the content with actual section paths if they exist.

    This function checks if the section exists in the build directory. If it does,
    it replaces the placeholder with the section's index path. If the section is
    optional and does not exist, it removes the placeholder from the content.

    Parameters
    ----------
    build_dir : str
        The directory where the documentation is being built.
    content : str
        The content string containing placeholders to be updated.
    section : str
        The name of the section to process.
    is_optional : bool, optional
        Whether the section is optional; defaults to False.
    log : logging.Logger, optional
        Logger for logging messages; defaults to the module logger.

    Returns
    -------
    str
        The content with placeholders replaced or removed.

    Raises
    ------
    FileNotFoundError:
        If non-optional section does not exist
    """
    section_path = os.path.join(build_dir, section, "index.rst")
    if not is_optional and not os.path.exists(section_path):
        if not (
            section == "releases" and os.path.exists(os.path.join(build_dir, section))
        ):  # need this second level check because releases/index.rst is auto-generated
            log.debug(f"Files in {os.path.join(build_dir, section)}")
            log.debug(os.listdir(os.path.join(build_dir, section)))
            raise FileNotFoundError(
                f"Required section {section} does not exist as {section_path}"
            )
    if is_optional and not os.path.exists(section_path):
        log.debug(f"Not adding optional section {section}")
        return content.replace(get_section_replace_string(section), "")
    if is_optional:
        log.debug(f"Including optional section {section}")
    return content.replace(get_section_replace_string(section), f"{section}/index")


def generate_top_level_index_file(
    build_dir,
    geoips_docs_dir,
    package_name,
    required_sections,
    optional_sections,
    log=logging.getLogger(__name__),
):
    """
    Create top-level index.rst file from template and available sections.

    This function reads the index template, replaces placeholders with section paths
    or removes them if they are optional and not present. It replaces the package name
    placeholder with the actual package name. The result is written to the index.rst
    file in the build directory.

    Parameters
    ----------
    build_dir : str
        Directory where the documentation is being built.
    geoips_docs_dir : str
        Path to the GeoIPS documentation dir.
    package_name : str
        The name of the package being documented.
    required_sections : list of str
        Sections that must be included in the index.
    optional_sections : list of str
        Sections that can be included if they exist.
    log : logging.Logger, optional
        Logger for logging messages; defaults to the module logger.
    """
    template_index_file_path = os.path.join(
        geoips_docs_dir, "source", "_templates", "index.template.rst"
    )
    build_index_file_path = os.path.join(build_dir, "index.rst")

    content = return_file_content(template_index_file_path)

    # Replace required sections
    # for section in required_sections:
    #    content = update_content_for_section(build_dir, content, section, log=log)

    # Replace optional sections
    for section in optional_sections:
        content = update_content_for_section(
            build_dir, content, section, is_optional=True, log=log
        )

    # Replace package name
    content = content.replace("PKGNAME", package_name)

    with open(build_index_file_path, "wt") as build_index_file:
        build_index_file.write(content)


def copy_template_files_to_non_geoips_repo(geoips_docs_dir, build_dir):
    """
    Copy static and template files to the build directory for non-GeoIPS packages.

    This function copies necessary static files, LaTeX styles, and templates from the
    GeoIPS documentation directory to the build directory to ensure consistency in
    documentation appearance for packages other than GeoIPS.

    Parameters
    ----------
    geoips_docs_dir : str
        The path to the GeoIPS documentation templates directory.
    build_dir : str
        The directory where the documentation is being built.
    """
    shutil.copytree(
        os.path.join(geoips_docs_dir, "source", "_static"),
        os.path.join(build_dir, "_static"),
    )
    template_path = os.path.join(build_dir, "_templates")
    os.makedirs(template_path, exist_ok=True)
    shutil.copy(
        os.path.join(geoips_docs_dir, "source", "_templates", "geoips_footer.html"),
        template_path,
    )


def create_conf_py_from_template(build_dir, geoips_docs_dir, package_name):
    """
    Generate the Sphinx configuration file (conf.py) from GeoIPS template.

    This function reads the template conf.py file, replaces the package name placeholder
    with the actual package name, and writes the result to the conf.py file in the build
    directory. The custom conf.py is used by Sphinx during the build process.

    Parameters
    ----------
    build_dir : str
        The directory where the documentation is being built.
    geoips_docs_dir : str
        The path to the GeoIPS documentation docs directory.
    package_name : str
        The name of the package being documented.
    """
    template_conf_file_path = os.path.join(
        geoips_docs_dir, "source", "_templates", "sphinx_conf.template.py"
    )
    build_conf_file_path = os.path.join(build_dir, "conf.py")

    with open(template_conf_file_path, "rt") as template_conf_file:
        with open(build_conf_file_path, "wt") as build_conf_file:
            for line in template_conf_file:
                build_conf_file.write(line.replace("PKGNAME", package_name))


def stage_docs_files_for_building(
    source_dir,
    package_name,
    geoips_docs_dir,
    build_dir,
    log=logging.getLogger(__name__),
):
    """
    Set up the documentation source files in the build directory before building.

    This function copies the package documentation source files into a build directory.
    If the package isn't 'geoips', it setups up template files from the GeoIPS docs.
    It then creates a customized conf.py and index.rst files tailored for the package.

    Parameters
    ----------
    source_dir : str
        The source directory containing the package's documentation files.
    package_name : str
        The name of the package being documented.
    geoips_docs_dir : str
        The path to the GeoIPS documentation templates.
    build_dir : str
        The directory where the documentation will be built.
    log : logging.Logger, optional
        Logger for logging messages; defaults to the module logger.
    """
    shutil.copytree(source_dir, build_dir)  # copy docs files over

    if not package_name == "geoips":
        copy_template_files_to_non_geoips_repo(geoips_docs_dir, build_dir)

    create_conf_py_from_template(build_dir, geoips_docs_dir, package_name)

    required_sections, optional_sections = get_sections(package_name)
    generate_top_level_index_file(
        build_dir,
        geoips_docs_dir,
        package_name,
        required_sections,
        optional_sections,
        log,
    )


def build_module_apidocs_with_sphinx(
    module_path, apidoc_build_path, log=logging.getLogger(__name__)
):
    """
    Generate API documentation RST files using Sphinx apidoc.

    This function constructs the necessary arguments for sphinx-apidoc to generate
    the API documentation RST files, excluding certain paths as specified. It then
    executes sphinx-apidoc and checks for successful completion. It calls the ``main``
    function of sphinx, and passes arguments to it. This is clunky, but after
    reviewing the sphinx codebase it seemed to be the best option. Further,
    there is no (or at least extremely little) documentation on how to use
    sphinx inside of python; hence the current approach.

    Parameters
    ----------
    module_path : str
        The path to the Python module dir for which to generate API docs.
    apidoc_build_path : str
        The output directory where the generated RST files will be placed.
    log : logging.Logger, optional
        Logger for logging messages; defaults to the module logger.

    Raises
    ------
    Exception
        If the sphinx-apidoc command fails.
    """
    arguments = [
        "--no-toc",
        "-o",  # flag for output path
        apidoc_build_path,  # output path
        module_path,  # module path
        "*/lib/*",  # exclude path
    ]
    log.debug(f"Running sphinx apidoc with arguments '{' '.join(arguments)}'")

    if not sphinx_apidoc(arguments) == 0:
        raise Exception("Sphinx API Doc build failed")


def build_docs_with_sphinx(build_dir, built_dir, log=logging.getLogger(__name__)):
    """
    Build documentation HTML files using Sphinx.

    See also docstring for build_module_apidocs_with_sphinx for information
    on why command line-like arguments are being passed to sphinx like this.

    Parameters
    ----------
    build_dir : str
        The source directory containing the documentation to be built.
    built_dir : str
        The destination directory where the built HTML files will be placed.
    log : logging.Logger, optional
        Logger for logging messages; defaults to the module logger.

    Notes
    -----
    This overrides the exception handler to re-raise exceptions, allowing for error
    handling inline.
    """
    arguments = [
        "-b",  # builder name
        "html",  # uses the html builder
        "-W",  # fail on warnings
        "-v",
        build_dir,  # folder to build from
        built_dir,  # folder to build to
    ]
    log.debug(f"Running sphinx build with arguments '{' '.join(arguments)}'")

    # for currying, prevents recursion error
    original_handler = sphinx_build_module.handle_exception

    def override_handle_exception(app, args, exc, error):
        original_handler(app, args, exc, error)
        raise exc  # raise exceptions so they bubble up in this script

    sphinx_build_module.handle_exception = override_handle_exception
    sphinx_build_module.main(arguments)  # build docs with sphinx


def build_release_note_from_dir_with_brassy(
    release_dir, release_filename, version, header_file, log=logging.getLogger(__name__)
):
    """
    Build release note file by combining YAML files in a specified directory.

    This function aggregates YAML files in the given `release_dir` directory
    and generates a release note in the specified `release_filename` file
    with the `brassy` build tool.

    Parameters
    ----------
    release_dir : str
        Path to the directory containing the YAML files to include in the
        release notes.
    release_filename : str
        Path to the output file where the generated release note will be saved.
    version : str
        Version string to label the release note content.
    header_file : str
        Path to a header file to include at the beginning of the release note.
    log : logging.Logger, optional
        Logger for debug information (default is a logger named after the
        module).

    Returns
    -------
    None
        This function does not return a value; it writes the release note to
        `release_filename`.

    Examples
    --------
    >>> build_release_note_from_dir_with_brassy(
    ...     release_dir="/path/to/yaml_files",
    ...     release_filename="release_notes.rst",
    ...     version="1.2.3",
    ...     header_file="header.rst"
    ... )
    This example writes the release notes from YAML files in `/path/to/yaml_files`
    to `release_notes.yaml`, as version "1.2.3" and includes the content
    of `header.md` as a header.
    """
    log.debug(
        f"Building yaml files in {release_dir} into {release_filename} using brassy"
    )
    with open(release_filename, "w") as f:
        release_note_content = brassy_build.build_release_notes(
            [release_dir],
            Console(),
            Progress().open,
            version=version,
            header_file=header_file,
        )
        f.write(release_note_content)


def build_release_notes_with_brassy(
    releases_dir,
    license_url,
    log=logging.getLogger(__name__),
    save_temp_dir=None,
):
    """Generate release notes for each subdirectory in a specified releases directory.

    Uses the `brassy` build tool and the `pinkrst` formatting tool to build release
    notes with a license disclaimer at the top.

    Parameters
    ----------
    releases_dir : str
        The path to the main directory containing individual release directories.
    license_url : str
        The URL pointing to the license or distribution statement for the release notes.
    log : logging.Logger, optional
        Logger instance used for logging debug and warning messages. By default,
        uses a logger with the module's name.
    save_temp_dir : str
        Optional path to directory to save temp files for reference in debugging

    Notes
    -----
    Each subdirectory in `releases_dir` is assumed to correspond to a release version.
    This function generates a header file containing a distribution statement that
    includes `license_url`. For each release directory, an `.rst` file with the
    release notes is created and processed by `build_release_note_from_dir_with_brassy`.
    The function will ignore directories named "upcoming" and log a warning.

    Warnings
    --------
    Directories in `releases_dir` named "upcoming" are skipped, and a warning logged.

    Raises
    ------
    FileNotFoundError
        If `releases_dir` does not exist or is not a directory.
    PermissionError
        If there are permissions issues accessing `releases_dir` or its subdirectories.

    Example
    -------
    >>> build_release_notes_with_brassy('/path/to/releases',
    >>>                                 'https://example.com/license')

    """
    release_dirs = filter(
        os.path.isdir,
        [os.path.join(releases_dir, rd) for rd in os.listdir(releases_dir)],
    )

    with tempfile.NamedTemporaryFile(mode="w") as header_file:
        header_file.write(
            "\n".join(
                [
                    ".. dropdown:: Distribution Statement",
                    "  | This file is auto-generated. Please abide by the license",
                    f"  | found at {license_url}.",
                ]
            )
        )

        for release_dir in release_dirs:
            release_filename = release_dir + ".rst"
            release_version = release_dir.replace(releases_dir + "/", "")
            if release_version == "upcoming":
                log.warning(
                    f"Skipping release dir {release_dir}"
                    "because it's version is 'upcoming'"
                )
            log.debug(f"Setting version of {release_filename} to {release_version}")
            build_release_note_from_dir_with_brassy(
                release_dir, release_filename, release_version, header_file.name
            )
            if save_temp_dir:
                log.info(f"Writing temp files to {save_temp_dir}")
                os.makedirs(save_temp_dir, mode=0o755, exist_ok=True)
                shutil.copy(release_filename, save_temp_dir)
            # TODO: pythonize and call directly; requires update to pink
            call(["pink", release_filename], shell=False)

    log.info(
        "Generating index.rst for release notes, "
        "this will be done by brassy in the future."
    )
    generate_release_note_index(os.path.join(releases_dir, "index.rst"), releases_dir)


def get_auxiliary_files():
    """
    Load auxiliary file information from a YAML configuration file.

    This function reads the `auxiliary_files.yaml` file located in the directory
    specified by `__location__` and returns the parsed dictionary of auxiliary
    files. The returned data corresponds to the "auxiliary files" field in the
    YAML file, which should be defined relative to the repository's root directory.

    Returns
    -------
    dict
        A dictionary containing information about auxiliary files as defined
        under the "auxiliary files" key in the YAML file.

    Examples
    --------
    >>> aux_files = get_auxiliary_files()
    >>> aux_files
    {'config.json': 'config/config.json', 'notes.txt': 'docs/notes.txt'}

    Notes
    -----
    The `__location__` variable must point to the directory containing the
    `auxiliary_files.yaml` file.
    """
    with open(os.path.join(__location__, "auxiliary_files.yaml"), "r") as f:
        data = yaml.safe_load(f)
    return data["auxiliary files"]  # relative to root of repo_dir


def import_non_docs_files(repo_dir, build_dir, log=logging.getLogger(__name__)):
    """
    Copy auxiliary non-documentation files from the repository to the build directory.

    Creates an 'import' subdirectory within the specified build directory and
    copies auxiliary files from the repository directory to this 'import' directory.

    Parameters
    ----------
    repo_dir : str
        Path to the repository directory containing the source files.
    build_dir : str
        Path to the build directory where the files will be copied.

    Raises
    ------
    FileNotFoundError
        If any of the auxiliary files do not exist in the repository directory.

    Notes
    -----
    The list of auxiliary files to copy is defined within the function.
    """
    auxiliary_files = get_auxiliary_files()
    import_dir = os.path.join(build_dir, "import")
    os.mkdir(import_dir)
    for file in auxiliary_files:
        filename = os.path.basename(file)
        source = os.path.join(repo_dir, file)
        dest = os.path.join(import_dir, filename)
        log.info(f"Copying {source} to {dest}")
        try:
            log.info(f"Copying {source} to {dest}")
            shutil.copyfile(source, dest)
        except FileNotFoundError:
            log.warning(f"Could not fine aux file {source}")


def build_html_docs(
    repo_dir,
    build_dir,
    geoips_docs_dir,
    package_name,
    output_dir,
    force_overwrite,
    license_url,
    log=logging.getLogger(__name__),
    save_temp_dir=None,
):
    """
    Build the HTML documentation for package.

    Sets up the documentation files, generates API documentation,
    builds the HTML documentation using Sphinx, and then copies the built files
    to the specified output directory if the build is successful.

    Parameters
    ----------
    repo_dir : str
        The root directory of the package's repository.
    build_dir : str
        The directory where the documentation source files will be staged.
    geoips_docs_dir : str
        The path to the GeoIPS documentation templates (usually geoips/docs).
    package_name : str
        The name of the package that docs are being built for.
    output_dir : str
        The directory where the final built documentation will be placed.
    force_overwrite : bool
        If true, replace output_dir if exists already.
    license_url : str
        URL that points to the license for the package.
    log : logging.Logger, optional
        Logger for logging messages; defaults to the module logger.
    save_temp_dir : str
        Optional path to directory to save temp files for reference in debugging
    """
    log.info("Setting docs files up for building")
    # copy and validate files
    stage_docs_files_for_building(
        os.path.join(repo_dir, "docs", "source"),
        package_name,
        geoips_docs_dir,
        build_dir,
        log=log,
    )

    # grab auxillary files not in docs and place them in "import" dir
    import_non_docs_files(repo_dir, build_dir, log=log)

    # build release rst files
    log.info("Building API docs")
    releases_dir = os.path.join(build_dir, "releases")
    build_release_notes_with_brassy(
        releases_dir, license_url, save_temp_dir=save_temp_dir
    )

    # build api doc rst files
    apidoc_build_path = os.path.join(build_dir, f"{package_name}_api")
    module_path = os.path.join(repo_dir, package_name)  # module path
    build_module_apidocs_with_sphinx(module_path, apidoc_build_path, log=log)

    # build final html files
    with tempfile.TemporaryDirectory() as built_dir:
        log.info("Building docs")
        build_docs_with_sphinx(build_dir, built_dir, log=log)
        log.debug("Docs built successfully")
        if os.path.exists(output_dir) and force_overwrite:
            log.info(
                f"Removing output directory {output_dir} in preparation of writing"
                "built docs"
            )
            shutil.rmtree(output_dir)
        try:
            shutil.copytree(built_dir, output_dir)
        except FileExistsError:
            raise FileExistsError(
                f"Output dir {output_dir} exists."
                " Re-running with --force will overwrite existing dir"
            )
        print(f"Docs built and written to {output_dir}")


def main(
    repo_dir,
    package_name,
    geoips_docs_dir,
    output_dir,
    force_overwrite,
    license_url,
    save_temp_dir,
):
    """Prepare for and execute documentation build.

    This function initializes logging, validates paths and package installation,
    logs the configuration, and runs the documentation build process within a temporary
    directory to avoid cluttering the filesystem.

    Parameters
    ----------
    repo_dir : str
        The root directory of the package's repository.
    package_name : str
        The name of the package to build documentation for.
    geoips_docs_dir : str
        The path to the GeoIPS documentation directory (usually geoips/docs)
    output_dir : str
        The directory where the built documentation will be placed.
    save_temp_dir : str
        Optional path to directory to save temp files for reference in debugging
    """
    log = init_logger(True)
    log.debug("Program initialized")

    validate_package_is_installed(package_name, logger=log)

    for path in (
        repo_dir,
        geoips_docs_dir,
        os.path.join(repo_dir, "docs"),
        os.path.join(repo_dir, package_name),
    ):
        validate_path_exists_and_is_git_repo(path, logger=log)

    log.info(f"Repo path is {repo_dir}")
    log.info(f"Repo docs path is {os.path.join(repo_dir, 'docs')}")
    log.info(f"GeoIPS docs path is {geoips_docs_dir}")
    log.info(f"Package name is {package_name}")

    with tempfile.TemporaryDirectory() as docs_build_dir_container:
        build_dir = os.path.join(docs_build_dir_container, "build")
        build_html_docs(
            repo_dir,
            build_dir,
            geoips_docs_dir,
            package_name,
            output_dir,
            force_overwrite,
            license_url,
            log=log,
            save_temp_dir=save_temp_dir,
        )


# Execute the main function with command line arguments
if __name__ == "__main__":
    args = parse_args_with_argparse()
    main(
        repo_dir=args.repo_path,
        package_name=args.package_name,
        geoips_docs_dir=args.geoips_docs_path,
        output_dir=args.output_dir,
        force_overwrite=args.force,
        license_url=args.license_url,
        save_temp_dir=args.save_temp_dir,
    )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS GEOTIFF filename formatter."""

# Python Standard Libraries
import logging

from geoips.filenames.base_paths import PATHS as gpaths

LOG = logging.getLogger(__name__)

interface = "filename_formatters"
family = "standard"
name = "geotiff_fname"


def call(
    area_def,
    xarray_obj,
    product_name,
    coverage=None,
    output_type="tif",
    output_type_dir=None,
    product_dir=None,
    product_subdir=None,
    source_dir=None,
    basedir=gpaths["GEOTIFF_IMAGERY_PATH"],
    output_dict=None,
):
    """GEOTIFF filename formatter.

    This uses the standard "geoips_fname" formatter, but with a default
    output type of "tif".
    """
    from geoips.interfaces import filename_formatters
    from geoips.sector_utils.utils import is_sector_type

    if is_sector_type(area_def, "tc"):
        fname_fmt_plugin = filename_formatters.get_plugin("tc_fname")
        basedir = gpaths["TCWWW"]
    else:
        fname_fmt_plugin = filename_formatters.get_plugin("geoips_fname")
    geotiff_fname = fname_fmt_plugin(
        area_def,
        xarray_obj,
        product_name,
        coverage=coverage,
        output_type=output_type,
        output_type_dir=output_type_dir,
        product_dir=product_dir,
        product_subdir=product_subdir,
        source_dir=source_dir,
        basedir=basedir,
        output_dict=output_dict,
    )

    return geotiff_fname



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test all Output Checker plugins."""
import pytest

from geoips.interfaces import output_checkers
from geoips.commandline import log_setup

log_setup.setup_logging()

# If this is set to a path, this path will be used for storing the test data files
# Otherwise a tmp_path directory will be used and deleted at the end of each test
# For more information see here: https://docs.pytest.org/en/7.1.x/how-to/tmp_path.html
permanent_test_data_dir = None


class TestOutputCheckers:
    """TestOutputChecker class, defining methods as well."""

    from geoips.interfaces import output_checkers

    available_output_checkers = [
        str(checker.name) for checker in output_checkers.get_plugins()
    ]

    def compare_plugin(self, tmp_path, plugin):
        """Test the comparision of two products with the appropriate Output Checker."""
        if permanent_test_data_dir:
            compare_paths, output_paths = plugin.module.get_test_files(
                permanent_test_data_dir
            )
        else:
            compare_paths, output_paths = plugin.module.get_test_files(tmp_path)
        plugin.module.perform_test_comparisons(plugin, compare_paths, output_paths)

    @pytest.mark.parametrize("checker_name", available_output_checkers)
    def test_plugins(self, tmp_path, checker_name):
        """Test all output_checkers that are ready for testing."""
        plugin = output_checkers.get_plugin(checker_name)
        # Note "long" running output checkers unit tests are not currently
        # supported.  For now, xfail if we come across a "long" unit test.
        # We will eventually implement "long" running unit tests, but for
        # not just xfail.
        if hasattr(plugin.module, "get_test_files_long") or not hasattr(
            plugin.module, "perform_test_comparisons_long"
        ):
            pytest.xfail(checker_name + " should be run with the long unit tests.")
        if not hasattr(plugin.module, "get_test_files") or not hasattr(
            plugin.module, "perform_test_comparisons"
        ):
            pytest.xfail(checker_name + " is not ready to be tested yet.")
        self.compare_plugin(tmp_path, plugin)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Full NetCDF output format, writing out all channels in full dataset."""
import logging

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "xarray_data"
name = "netcdf_full"


def call(xarray_obj, product_names, output_fnames, clobber=False):
    """Write GeoIPS style NetCDF to disk."""
    from geoips.plugins.modules.output_formatters.netcdf_xarray import (
        write_xarray_netcdf,
    )

    for ncdf_fname in output_fnames:
        write_xarray_netcdf(xarray_obj, ncdf_fname, clobber=clobber)
    return output_fnames



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""NASA SPoRT Dust RGB product.

Data manipulation steps for the "nasa_dust_rgb" product.
This algorithm expects Brightness Temperatures in units of degrees Kelvin
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "nasa_dust_rgb"


def call(arrays):
    """nasa_dust_rgb product algorithm data manipulation steps.

    This algorithm expects Brightness Temperatures in units of Kelvins,
    and returns red green and blue gun arrays.

    Parameters
    ----------
    data : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of channels list above
        * Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output

    Channels: RED: 15-13 (-6.7,2.6,G1);
              GRN: 14-11 (-0.5,20.0,G2.5)
              BLU: 13 (-11.95,15.55,G1)

    """
    c11 = arrays[0]
    c13 = arrays[1]
    c14 = arrays[2]
    c15 = arrays[3]

    from geoips.data_manipulations.corrections import apply_data_range, apply_gamma

    red = c15 - c13
    red = apply_data_range(
        red,
        -6.7,
        2.6,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    red = apply_gamma(red, 1.0)

    grn = c14 - c11
    grn = apply_data_range(
        grn,
        -0.5,
        20.0,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    grn = apply_gamma(grn, 2.5)

    blu = c13 - 273.15
    blu = apply_data_range(
        blu,
        -11.95,
        15.55,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    blu = apply_gamma(blu, 1.0)

    from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)
    return rgba



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Filename formatters interface module."""

from geoips.interfaces.base import BaseModuleInterface


class FilenameFormattersInterface(BaseModuleInterface):
    """Specification for formatting the full path and file name.

    File path and name formatting is determined using attributes within the
    GeoIPS xarray objects.
    """

    name = "filename_formatters"

    required_args = {
        "standard": ["area_def", "xarray_obj", "product_name"],
        "xarray_metadata_to_filename": ["xarray_obj"],
        "data": ["area_def", "xarray_obj", "product_names"],
        "standard_metadata": ["area_def", "xarray_obj", "product_filename"],
        "xarray_area_product_to_filename": ["xarray_obj", "area_def", "product_name"],
    }
    required_kwargs = {
        "standard": [
            "coverage",
            "output_type",
            "output_type_dir",
            "product_dir",
            "product_subdir",
            "source_dir",
            "basedir",
        ],
        "xarray_metadata_to_filename": ["extension", "basedir"],
        "data": [
            "coverage",
            "output_type",
            "output_type_dir",
            "product_dir",
            "product_subdir",
            "source_dir",
            "basedir",
        ],
        "standard_metadata": ["metadata_dir", "metadata_type", "basedir"],
        "xarray_area_product_to_filename": ["output_type", "basedir", "extra_field"],
    }

    # The functions below were commented out as they included errors, and were not used
    # by GeoIPS at this time. 9/27/23

    # def find_duplicates(self, *args, **kwargs):
    #     """Find duplicate files."""
    #     try:
    #         func = self.get_plugin_attr(name, "find_duplicates")
    #     except AttributeError:
    #         raise AttributeError(
    #             f'Plugin {name} does not have a "find_duplicates" function.'
    #         )

    #     duplicates = func()

    # def remove_duplicates(self):
    #     """Remove duplicate files."""
    #     duplicates = self.find_duplicates()


filename_formatters = FilenameFormattersInterface()
